/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
 
/*
 *
 * Copyright (c) 1996 Brian Maso. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and
 * without fee is hereby granted.
 *
 *
 * Written By: Brian Maso
 * Date: 2/29/96
 * Description: Scheduler class, which is
 * part of the MultiSynch package of synchronization.
 *
 * Usage Notes:
 *
 * There is only a single instance of the Scheduler class.
 * All the public methods of the Scheduler class are declared
 * static to ensure this constraint. The constructor for the
 * class has no access declaration, so it is public to the
 * other members of the maso.synch package (ie, any Synch
 * object). The Synch class constructor is designed to
 * create the only Scheduler instance as part of the first
 * Synch object's construction.
 *
 * The Scheduler object manages two hashes:
 * + The hash of Synch, where a Vector of the
 *   Threads waiting for the Synch is the contents.
 * + The hash of Threads, where the Vector of
 *   Synch being waited for is the contents.
 *
 * Each Synch registers itself with the Scheduler in its
 * creator. The Synch is added to the hash of Synchs with an
 * association to an empty Vector.
 *
 * To wait for one or more Synch objects, a Thread calls waitFor(),
 * passing either a single Synch or a Vector of Synch objects.
 * A copy of the Vector is made (or a new one created) and
 * stored under the Thread in _hashThreadToSynchs. Backwards
 * entries under each Synch are made in _hashSynchToThreads.
 *
 * A single Thread runs using Scheduler.run(). This is termed the
 * _threadScheduler. It waits indefinitely for the addSchedulerItem
 * method to be called. A call to addSchedulerItem indicates that
 * either a Synch has become unsignalled (and associated Threads
 * waiting for that Synch need to be tested to see if they're
 * to be released), or a Thread has started waiting for one or
 * more Synchs (so the Synchs must be tested immediately to see
 * if the Thread needs to wait at all).
 *
 * Each Thread waiting for one or more Synchs is suspended using
 * a ThreadSuspender object. The ThreadSuspender class has two
 * methods: suspendCurrentThread() and resumeWaitingThread(). The
 * Scheduler object's _threadScheduler calls resumeWaitingThread()
 * to wake up the thread waiting in suspendCurrentThread().
 *
 * When a Thread is released, then it is deleted from both
 * internal Hashtables _hashThreadToSynchs and _hashSynchToThreads,
 * since a reference to it is no longer required.
 * 
 */

package com.radview.util.synch;

import java.util.*;
import java.lang.*;

	/**
	 * The Scheduler is a single-instance object which is
	 * the heart of thread synchronization in the maso.synch
	 * package. To synchronize a thread on the state of
	 * a Synch object, use Scheduler.waitFor(), like so:
	 *
	 * <PRE>
	 *	public class RunnableObject implements Runnable {
	 *		Thread _background = new Thread(this);
	 *		Signal _signal = new Signal(unsignaled);
	 *
	 *		public RunnableObject() {
	 *			_background.start();
	 *		}
	 *
	 *		public void run() {
	 *			Scheduler.waitFor(_signal);
	 *			// Do something...guaranteed that _signal
	 *			// has changed its state to "signaled".
	 *		}
	 *	}
	 *
	 * </PRE>
	 *
	 * Overloaded versions of <I>waitFor()</I> exist so a thread
	 * can synchronize on the state of multiple Synch objects.
	 * The thread can synchronize itself on all or any one of
	 * the multiple Synch objects.
	 *
	 * @see Synch
	 *
	 * @version 1.0
	 * @author Brian Maso
	 */

public class Scheduler implements Runnable {
	Hashtable _hashThreadToSynchs = new Hashtable();
	Hashtable _hashSynchToThreads = new Hashtable();
	Hashtable _hashThreadToSuspender = new Hashtable();
	Vector _vectSchedulerItems = new Vector();
	Thread _threadScheduler;
	static Scheduler _scheduler = null;
	
	static String strCopyright =
			"maso.synch package initialized.\n\tCopyright (C) 1996, Brian Maso. All rights reserved.\n\tLicensed for private use or for public NON-COMMERCIAL use.\n\thttp://www.europa.com/~bmaso/java/synch";

	/**
	 * Constant AND is an optional flag for <I>waitFor</I>.
	 * It is used in conjunction with a list of Synch objects. When
	 * specified the Scheduler will suspend the synchronized thread
	 * until the states of all Synchs in the list are "signaled".
	 *
	 * @see Scheduler#waitFor
	 */
	public static final int AND = 0;
	
	/**
	 * Constant OR is an optional flag for <I>waitFor</I>.
	 * It is used in conjunction with a list of Synch objects. When
	 * specified the Scheduler will suspend the synchronized thread
	 * until the state of at least one of the Synchs in the list is "signaled".
	 */
	public static final int OR  = 1;

	/**
	 * Constant ABANDONED is the value returned by <I>waitFor()</I>
	 * if any one of the Synch objects a synchronized thread is
	 * waiting for are destroyed.
	 *
	 * @see Scheduler#waitFor
	 */
	public static final int ABANDONED = -2;

	/**
	 * Constant TIMEDOUT is the value returned by <I>waitFor()</I>
	 * if the time limit specified to <I>waitFor()</I> expires before a
	 * synchronized thread is released.
	 *
	 * NOT YET IMPLEMENTED. To get the same functionality, use a Timer
	 * object.
	 *
	 * @see Scheduler#waitFor
	 * @see Timer
	 */
	public static final int TIMEDOUT  = -1;

	/**********
	 * Constructor is private so nobody can use the
	 * new operator to create a new Scheduler. Also,
	 * only a single instance of the Scheduler should ever
	 * exist, so constructor ensures this is so. To get
	 * a reference to the single Scheduler instance, use
	 * static Scheduler.getScheduler().
	 **********/
	private Scheduler() {
		if(null != _scheduler)
			throw(new IllegalArgumentException(
					"Only a single instance of the Scheduler is allowed!"));
					
		_scheduler = this;
		_threadScheduler = new Thread(this, "Scheduler");
		_threadScheduler.setDaemon(true);
		_threadScheduler.start();
		
		/* Copyright and identification message so that the world
		 * can see this is a copyrighted package.
		 */
//		System.out.println(strCopyright);
		
		return;
	}
	
	static Scheduler getScheduler() {
		if(null == _scheduler)
			new Scheduler();
			
		return _scheduler;
	}
	
	/********************
	 * All external objects can only use these static methods
	 * to affect the single Scheduler object. This ensures
	 * only a single instance will even exist, it also makes
	 * things easier to read externally. That is, instead
	 * of "Scheduler.getScheduler().register(s)", you have
	 * "Scheduler.register(s)", a lot simpler to read.
	 *
	 * See awt.java.Toolkit for an example of another single-
	 * instance object. I didn't like how they did it, so I'm
	 * doing it this way.
	 *
	 * See private i_<method-name>() method to see what static
	 * method <method-name>() does.
	 ********************/
	
	static void register(Synch s) {
		getScheduler().i_register(s);
	}
	
	static void unregister(Synch s) {
		getScheduler().i_unregister(s);
	}
	
	/**
	 * Synchronizes the calling thread on the state of the Synch
	 * object passed. The calling thread will be suspended until
	 * the state of Synch <I>s</I> changes to "signaled".
	 * When the state of the Synch is "signaled",
	 * the calling thread will be released.
	 *
	 * @param	s		The Synch object to wait for
	 *
	 * @return			A return value of <I>Scheduler.ABANDONED</I> means
	 *				that Synch <I>s</I> was destroyed
	 *				before its state became "signaled".
	 *				Zero is returned otherwise.
	 *
	 * @see Synch
	 */
	
	public static int waitFor(Synch s) {
		return getScheduler().i_waitFor(s);
	}
	
	/**
	 * Synchronizes the calling thread on the state of one or more Synch
	 * objects passed in a Vector. The calling thread will be suspended until
	 * the state of all the Synchs in Vector <I>v</I> changes to "signaled".
	 * When the state of the Synchs are "signaled", the calling thread will be released.
	 *
	 * @param	v		A Vector filled with Synch objects to wait for.
	 *
	 * @return			A return value of <I>Scheduler.ABANDONED</I> means
	 *				that one of the Synchs in <I>v</I> was destroyed
	 *				before its state became "signaled".
	 *				Zero is returned otherwise.
	 *
	 * @exception	java.lang.IllegalArgumentException	An
	 *				IllegalArgumentException is thrown if an object of
	 *				any other class than Synch occurs in the Vector
	 *				<I>v</I>.
	 *
	 * @see Synch
	 */
	public static int waitFor(Vector v) {
		return waitFor(v, AND);
	}
	
	/**
	 * Synchronizes the calling thread on the state of one or more Synch
	 * objects passed in a Vector. According to the value of <I>nflgAndOr</I>,
	 * the calling thread will either be suspended until all or any one of the Synchs
	 * in Vector <I>v</I> become signaled.
	 *
	 * @param	v		A Vector filled with Synch objects to wait for.
	 * @param	nflgAndOr	<UL>Has one of two values:
	 *				<LI>Scheduler.AND - thread is suspended until all
	 *				Synch elements in <I>v</I> become signaled.
	 *				<LI>Scheduler.OR - thread is suspended until at least one
	 *				of the Synch elements in <I>v</I> become signaled.
	 *				</UL>
	 *
	 * @return			A return value of <I>Scheduler.ABANDONED</I> means
	 *				that one of the Synchs in <I>v</I> was destroyed
	 *				before its state became "signaled".
	 *				The index of the first Synch in <I>v</I> to become
	 *				signaled is returned otherwise.
	 *
	 * @exception	java.lang.IllegalArgumentException	An
	 *				IllegalArgumentException is thrown if an object of
	 *				any other class than Synch occurs in the Vector
	 *				<I>v</I>.
	 *
	 * @see Synch
	 */
	public static int waitFor(Vector v, int nflgAndOr) {
		return getScheduler().i_waitFor(v, nflgAndOr);
	}
	
	static void notifysignaled(Synch s) {
		getScheduler().i_notifysignaled(s);
	}
	
	static void notifyUnsignaled(Synch s) {
		getScheduler().i_notifyUnsignaled(s);
	}
	
	/**********
	 * void i_register(Synch s)
	 *
	 * Registers the Synch by adding an empty Vector
	 * to the internal _hashSynchToThreads hash.
	 * If Synch is already registered, this call throws
	 * an IllegalArgumentException, since that should
	 * never happen anyway.
	 *
	 **********/
	
	private synchronized void i_register(Synch s) {
		// Is it already registered?
		if(null != _hashSynchToThreads.get(s))
			throw(new IllegalArgumentException(
					"Synch " + s + " is already registered!"));
					
		// Add Synch and a new empty Vector to the hash.
		_hashSynchToThreads.put(s, new Vector());
		
		return;
	}
	
	/**********
	 * void i_unregister(Synch s)
	 *
	 * Unregisters the Synch with this Scheduler. This is
	 * done as a result of Synch.destroy(). All Synchs should
	 * be destroyed, or else they have a reference in the
	 * Scheduler forever. If Synch still has Threads waiting
	 * for it, the Threads are all released with an ABANDONED
	 * return value. If the Synch is not registered, an
	 * IllegalArgumentException will be thrown, since this
	 * should never happen anyway.
	 *
	 **********/
	
	private synchronized void i_unregister(Synch s) {
		// Get Vector for this Synch. If none, then Synch
		// has been improperly registered, and an
		// IllegalArgumentException is thrown.
		Vector v = (Vector)_hashSynchToThreads.get(s);
		if(null == v)
			throw(new IllegalArgumentException(
					"Synch " + s + " has not been registered properly!"));
					
		// Get Enumeration for v. For each Thread in it,
		// forget it with an ABANDONED flag using
		// abandonThread().
		Enumeration e = v.elements();
		while(e.hasMoreElements()) {
			Thread t = (Thread)e.nextElement();
			abandonThread(t);
		}
		
		// Remove Synch s from internal hash.
		_hashSynchToThreads.remove(s);
		
		return;
	}
	
	/**********
	 * int waitFor(Synch s)
	 *
	 * The calling Thread wants to wait until Synch s becomes
	 * signaled. Accomplished by adding the Synch to the
	 * list of items the internal thread is waiting for.
	 * Thread waits, using wait(), indefinitely
	 * until Scheduler wakes for one of two reasons: Synch is
	 * or becomes signaled, or Synch is destroyed. Mostly this
	 * method is carried out using the overloaded Vector
	 * version of this same method.
	 *
	 **********/
	
	private int i_waitFor(Synch s) {
		// Create a Vector to store Synch s. Add
		// s to it.
		Vector v = new Vector();
		v.addElement(s);
		
		// Return whatever waitFor(Vector) returns.
		return waitFor(v);
	}
	
	/**********
	 * int waitFor(Vector)
	 *
	 * Waits for each Synch in the Vector. Performs an AND wait.
	 * That is, it waits for all the elements in the list to
	 * be signaled at the same time.
	 *
	 **********/
	
	private int i_waitFor(Vector v) {
		return waitFor(v, AND);
	}
	
	/**********
	 * int waitFor(Vector, int)
	 *
	 * Waits for each Synch in the Vector. Performs either an
	 * AND or an OR wait, according to nflgAndOr parameter (if
	 * neight AND or OR is value of that flag,
	 * IllegalArgumentException is thrown).
	 *
	 * If current Thread is already waiting, there has been a big
	 * problem, and we must throw an IllegalArgumentException.
	 *
	 * Thread waits indefinitely, using wait(). It is awoken
	 * when Synchs have been signaled. Return value Integer
	 * holds the return value, which is either the index of
	 * the Synch which was signaled (for OR), 0 for AND,
	 * ABANDONED, or TIMEDOUT (TBD).
	 *
	 * There are three steps to this method:
	 * 1. Add current Thread to internal storage devices of
	 *    the Scheduler. This must be synchronized.
	 * 2. Put the Thread to sleep in its ThreadSuspender. This
	 *    must not be synchronized since its the ThreadSuspender's
	 *    monitor which will be used. If Thread owns Scheduler's
	 *    monitor, then no other thread would be able to enter
	 *    synchronized methods while the current Thread waits
	 *    for the ThreadSuspender's monitor.
	 * 3. Remove current Thread from internal storage devices.
	 *    This must also be synchronized.
	 *
	 **********/

	
	private int i_waitFor(Vector v, int nflgAndOr) {
		Vector vectParams;
		ThreadSuspender ts;
		Thread t;
		
		// Step 1: add current Thread to internal storage
		//   devices of the Scheduler. This must be
		//   synchronized.
		synchronized(this) {
			// Get current Thread. If is already waiting, then
			// throw IllegalArgumentException.
			t = Thread.currentThread();
			if(null != _hashThreadToSynchs.get(t))
				throw(new IllegalArgumentException(
						"Thread " + t + " is already waiting!"));
					
			// Make sure nflgAndOr is either AND or OR. If not,
			// throw IllegalArgumentException.
			if((nflgAndOr != AND) && (nflgAndOr != OR))
				throw(new IllegalArgumentException(
						"Wait flags " + nflgAndOr + " not recognized!"));
					
			// For each Synch, modify its list of waiting
			// threads by adding this thread to it. Run through
			// list first to ensure all elements are Synchs.
			// Also, check to ensure the Synch is registered. If
			// not, then Synch has been destroyed and Thread should
			// get an ABANDONED return value.
			Enumeration e = v.elements();
			while(e.hasMoreElements()) {
				Object o = e.nextElement();
				if(!(o instanceof Synch))
					throw(new IllegalArgumentException(
							"Can only wait on Synch objects!"));
				if(null == _hashSynchToThreads.get((Synch)o))
					return ABANDONED;
			}
						
			e = v.elements();
			while(e.hasMoreElements()) {
				Synch s = (Synch)e.nextElement();
				Vector sv = (Vector)_hashSynchToThreads.get(s);
				sv.addElement(t);
			}			
		
			// Create Integer to hold return value, Integer
			// to hold nflgAndOr, and Vector to hold these two
			// Integers, called vectParams. The list of Synchs
			// associated with the Thread will have this vectParams
			// as its first element.
			Integer retObj = new Integer(TIMEDOUT);
			Integer flgObj = new Integer(nflgAndOr);
			vectParams = new Vector();
			vectParams.addElement(retObj);
			vectParams.addElement(flgObj);
		
			Vector vect = (Vector)v.clone();
			vect.insertElementAt(vectParams, 0);
		
			// Add entry for this Thread to hash.
			_hashThreadToSynchs.put(t, vect);
		
			// Create ThreadSuspender for this Thread, and
			// add it to _hashThreadToSuspender and internal
			// list of items to be looked at.
			ts = new ThreadSuspender();
			_hashThreadToSuspender.put(t, ts);
			addSchedulerItem(ts);
		}
		
		// Step 2: non-synchronized, suspend current Thread
		// in the ThreadSuspender.
		ts.suspendCurrentThread();
		
		// Step 3: Remove references to this Thread from
		// various lists. This must be synchronized.
		synchronized(this) {
			// Remove references to this Thread in hashes using
			// removeThreadRefs().
			removeThreadRefs(t);
			
			// Get rid of any reference to the ThreadSuspender.
			_hashThreadToSuspender.remove(t);
		
			// Now that this Thread has been awakenned, either
			// by SchedulerThread, or by timing out. Return
			// value is in first element of vectParams.
			Integer retObj = (Integer)vectParams.elementAt(0);
			return retObj.intValue();
		}		
	}
	
	/**********
	 * void notifysignaled(Synch)
	 *
	 * Called to notify the Scheduler that the state of a Synch
	 * has changed from unsignaled to signaled. Each Thread
	 * waiting for this Synch must be checked using a single
	 * SchedulerThread made for this Synch.
	 *
	 **********/
	
	private synchronized void i_notifysignaled(Synch s) {
		// add this Synch to the list of items for the
		// Scheduler to check, using addSchedulerItem();
		addSchedulerItem(s);
		return;
	}
	
	/**********
	 * void notifyUnsignaled(Synch)
	 *
	 * Removes any references to the Synch already in the
	 * queue. This prevents useless scheduling.
	 *
	 **********/
	
	private synchronized void i_notifyUnsignaled(Synch s) {
		// If Synch already in queue, get rid of it.
		removeSchedulerItem(s);
		return;
	}
	
	/**********
	 * void run()
	 *
	 * Only run by the Scheduler Thread.
	 *
	 * The Scheduler thread waits for an item to check. Items
	 * are accessed by getNextSchedulerItem(), which blocks
	 * until another item has been added using addSchedulerItem().
	 *
	 * Once item is found, all associated Threads are checked
	 * to see if they can be awoken. If so, then they are.
	 *
	 **********/
	
	public synchronized void run() {
		while(true) {
			// blocks until an item is available.
			Object o = getNextSchedulerItem();
			
			if(o instanceof ThreadSuspender) {
				ThreadSuspender ts =
						(ThreadSuspender)o;
				Thread t = ts.getSuspendedThread();
				if(gainThreadSynchs(t))
					resumeThread(t);
			}
		
			// else, item is a Synch object...
			else if(o instanceof Synch) {
				Synch s = (Synch)o;
				// For each Thread waiting for this Synch, try
				// to wake it after gaining all Synchs.
				Vector v =
					(Vector)_hashSynchToThreads.get(s);
				if(null == v)
					// Leftover SchedulerThread refering to
					// a Synch which has been unregistered. Ignore.
					return;
				Enumeration e = v.elements();
				while(e.hasMoreElements()) {
					Thread t = (Thread)e.nextElement();
					if(gainThreadSynchs(t))
						resumeThread(t);
				}
			}
		}
	}
	
	/**********
	 * abandonThread(Thread t)
	 *
	 * At least one Synch the Thread was waiting for was destroyed.
	 * Here we remove all references to the Thread in both
	 * lists and set the return value for the Thread to
	 * ABANDONED.
	 *
	 **********/
	
	private void abandonThread(Thread t) {
		Vector v = (Vector)_hashThreadToSynchs.get(t);
		Vector vectParams = (Vector)v.elementAt(0);
		
		// remove old placeholder for the return value,
		// and place in new one.
		vectParams.removeElementAt(0);
		vectParams.insertElementAt(new Integer(ABANDONED), 0);
		
		// Wake Thread before returning.
		resumeThread(t);
		
		return;
	}
	
	/**********
	 * void removeThreadRefs(Thread t)
	 *
	 * Removes all references to the thread on either hash.
	 *
	 **********/
	
	private void removeThreadRefs(Thread t) {
		// Remove any ref from threads hash. Store the
		// Vector of Synch so we know which Synchs to remove
		// refs from.
		Vector v = (Vector)_hashThreadToSynchs.get(t);
		_hashThreadToSynchs.remove(t);
		
		// Get rid of first entry in v, which holds return
		// value for wait, and there is a reference still
		// in a local variable in waitFor().
		v.removeElementAt(0);
		
		// Rest of elements are Synchs with associations to
		// the thread to get rid of.
		Enumeration e = v.elements();
		while(e.hasMoreElements()) {
			// Note that Synch s may have been unregistered,
			// in which case the Thread was abandoned and we
			// don't have to remove refs for that Synch, because
			// the Synch no longer exists.
			Synch s = (Synch)e.nextElement();
			Vector v2 = (Vector)_hashSynchToThreads.get(s);
			if(null != v2) {
				int i = v2.indexOf(t);
				if(-1 != i)
					v2.removeElementAt(i);
			}
		}

		return;
	}
	
	/**********
	 * boolean gainThreadSynchs(Thread t)
	 *
	 * Attempts to reserve and gain all the Synchs associated
	 * with Thread t. Returns true if that was possible. Note
	 * that it is possible for Thread t to not have refs in the
	 * hashes if it has already timed out.
	 *
	 **********/
	
	private boolean gainThreadSynchs(Thread t) {
		// Get Vector for the Thread.
		Vector v = (Vector)_hashThreadToSynchs.get(t);

		if(null == v) {
			// SchedulerThread leftover after a Thread
			// has been destroyed. Ignore.
			return false;
		}

		Enumeration e = v.elements();
		
		// First element is a Vector, the second element of
		// which holds an Integer which is either AND or OR,
		// indicating which way to attempt to get the Synchs.
		Vector vectParams = (Vector)e.nextElement();
		Integer nflgAndOrObj = (Integer)vectParams.elementAt(1);
		
		// Attempt to get Synchs differently base on type of
		// nflgAndOrObj.
		if(AND == nflgAndOrObj.intValue()) {
			boolean fAllReserved = true;
			int count;
			for(count=0 ; e.hasMoreElements() ; count++) {
				Synch s = (Synch)e.nextElement();
				if(!s.reserve(t)) {
					fAllReserved = false;
					break;
				}
			}
			
			e = v.elements();
			e.nextElement();

			// If fAllReserved, then go ahead and gain() all
			// the Synchs.
			if(fAllReserved) {
				while(e.hasMoreElements()) {
					Synch s = (Synch)e.nextElement();
					s.gain(t);
				}
				
				// Store a wait value of 0, indicating
				// success.
				vectParams.removeElementAt(1);
				vectParams.insertElementAt(new Integer(0), 1);

				return true;
			}
					
			// We weren't able to reserve all the Synchs, so
			// unreserve all of them.
			for( ; count>0 ; count-- ) {
				Synch s = (Synch)e.nextElement();
				s.unreserve(t);
			}
		}
		
		// Else, we're doing an OR search.
		else {
			// The first Synch which can be reserved causes
			// success. Store index so we know what return
			// value should be.
			boolean fSuccess = false;
			int i;
			for( i=0 ; e.hasMoreElements() ; i++ ) {
				Synch s = (Synch)e.nextElement();
				if(s.reserve(t)) {
					fSuccess = true;
					s.gain(t);
					break;
				}
			}
			
			// If successful, store wait return value and
			// return true.
			if(fSuccess) {
				vectParams.removeElementAt(0);
				vectParams.insertElementAt(new Integer(i), 0);
				
				return true;
			}
		}

		return false;
	}
	
	/**********
	 * void resumeThread(Thread t)
	 *
	 * Called by a SchedulerThread, the job of this
	 * method is to wake up Thread t, which has been
	 * suspended within a ThreadSuspender. The
	 * ThreadSuspender can be looked up in the Hashtable
	 * _hashThreadToSuspender under the Thread.
	 *
	 **********/
	
	private void resumeThread(Thread t) {
		ThreadSuspender ts =
				(ThreadSuspender)_hashThreadToSuspender.get(t);
		if(null == ts)
			// Leftover SchedulerThread after a Thread has been
			// destroyed. Ignore.
			return;
			
		ts.resumeWaitingThread();
		return;
	}
	
	/**********
	 * void addSchedulerItem(Thread t)
	 * void addSchedulerItem(Synch s)
	 *
	 * Adds the item to the private _vectSchedulerItems
	 * storage device. When added, notify() is called to
	 * wake up the scheduler thread, in case its waiting
	 * for an item in getNextSchedulerItem().
	 *
	 * Also, if a reference to the scheduler item already
	 * exists in the queue, then this method simply returns.
	 *
	 **********/
	
	void addSchedulerItem(Thread t) {
		addSchedulerItem((Object) t);
	}
	
	void addSchedulerItem(Synch s) {
		addSchedulerItem((Object) s);
	}
	
	private void addSchedulerItem(Object o) {
		if(-1 != _vectSchedulerItems.indexOf(o))
			return;
			
		_vectSchedulerItems.addElement(o);
		notify();
		return;
	}
	
	/**********
	 * void getNextSchedulerItem()
	 *
	 * Gets the next item in the _vectSchedulerItems storage
	 * device. Blocks until there are any. Note that the
	 * accomanying notify() call is in addSchedulerItem().
	 *
	 **********/
	
	private Object getNextSchedulerItem() {
		while(0 == _vectSchedulerItems.size()) {
			try {
				wait();
			} catch (Exception e) {}
		}
		Object o = _vectSchedulerItems.elementAt(0);
		_vectSchedulerItems.removeElementAt(0);
		return o;
	}
	
	/**********
	 * void removeSchedulerItem(Object o)
	 *
	 * If a reference to the object exists in the queue,
	 * then get rid of it.
	 *
	 **********/
	
	private void removeSchedulerItem(Object o) {
		int i;
		while(-1 != (i = _vectSchedulerItems.indexOf(o)))
			_vectSchedulerItems.removeElementAt(i);
	}
}

/********************
 * ThreadSuspender
 *
 * The very simple class is used by the Scheduler to put
 * Threads to sleep and wake them up at appropriate times.
 * A single ThreadSuspender is used to put a single Thread
 * to sleep and to wake it up.
 *
 ********************/

class ThreadSuspender {
	Thread _threadSuspended = null;
	
	public ThreadSuspender() {}
	
	public synchronized void suspendCurrentThread() {
		_threadSuspended = Thread.currentThread();
		if(null ==
				Scheduler.getScheduler()._hashThreadToSynchs.get(
						_threadSuspended)) {
			return;
		}
			
		try {
			notifyAll();
			wait();
		} catch (Exception e) {}
		
		return;
	}
	
	public synchronized void resumeWaitingThread() {
		try {
			while(null == _threadSuspended)
				wait();
		
			notifyAll();
		} catch (Exception e) {}
		
		return;
	}
	
	public synchronized Thread getSuspendedThread() {
		try {
			while(null == _threadSuspended)
				wait();
		} catch (Exception e) {}
		return _threadSuspended;
	}
}
