/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
 
/*
 *
 * Copyright (c) 1996 Brian Maso. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and
 * without fee is hereby granted. 
 *
 * The classic Semaphore: a synchronization object which is
 * signaled as long as there are no more than N owners. When
 * the N-th Thread gains ownership, the Semaphore becomes
 * unsignaled. When one of the N owners releases ownership,
 * the Semaphore becomes unsignaled.
 *
 */

package com.radview.util.synch;

/**
 * The classic Semaphore synchronization object. A Semaphore is an owned
 * Synch object which can be owned by up to N threads at a time. While
 * less than N threads have ownership of the Semaphore, the Semaphore is
 * signaled. Once the N-th thread gains ownership of the Semaphore, it
 * switches state to unsignaled. The Semaphore will again become signaled
 * when one of the N threads releases ownership.
 *
 * <P>A thread is automatically granted ownership of a Semaphore when it
 * waits for it and is released by the Scheduler. To release ownership of
 * the Semaphore, the thread must call <I>Semaphore.release()</I>. Note
 * that if a thread owning a Semaphore dies, then the ownsership can never be
 * released without destroying the Semaphore itself. This has the practicle
 * effect of reducing N by 1.
 *
 * <P>Note that an owner thread will immediately be released by
 * <I>Scheduler.waitFor()</I>, even if the Semaphore is unsignaled. All owned
 * Synch objects have this functionality.
 *
 * @version 1.0
 * @author Brian Maso
 */


public class Semaphore extends Synch {
	int _maxOwners;

	/**
	 * Specify the maximum number of owners to the Semaphore's
	 * constructor. The Seamphore will remain signaled until
	 * <I>maxOwners</I> threads gain ownership of the Semaphore
	 * using <I>Scheduler.waitFor()</I>.
	 *
	 * @param	maxOwners		The number of owners of the Semaphore.
	 */
	public Semaphore(int maxOwners) {
		super(signaled);
		_maxOwners = maxOwners;
		return;
	}
	
	/* *********
	 * boolean reserve(Thread t)
	 *
	 * This overrider for the Synch.reserve() method ensures
	 * the reservation does not make the number of owners
	 * greater than _maxOwners. Furthermore, if after the
	 * reservation there are _maxPwners owners, then the
	 * state of the Semaphore changes to reserved.
	 *
	 ********* */
	
	boolean reserve(Thread t) {
		// First, is Thread already an owner or reserver?
		if(null != _hashOwnership.get(t))
			return true;
			
		// Is count already too high?
		if(_hashOwnership.size() >= _maxOwners)
			return false;
			
		// Let parent class reserve space.
		super.reserve(t);
		
		// If count is now up to max, then go to
		// reserved state.
		if(_hashOwnership.size() == _maxOwners)
			setState(reserved);
			
		return true;
	}
	
	/* *********
	 * void gain(Thread t)
	 *
	 * After performing parent class gain(), if count
	 * of owners is up to max, then set state to unsignaled.
	 *
	 ********* */
	
	void gain(Thread t) {
		super.gain(t);
		if(_hashOwnership.size() == _maxOwners)
			setState(unsignaled);
		return;
	}
	
	/* *********
	 * void release()
	 *
	 * Must be called within the context of an owner Thread.
	 * If not, an IllegalArgumentException is thrown. If count
	 * of owners after release is less than max, set state to
	 * signaled.
	 *
	 ********* */
	
	/**
	 * Use <I>release()</I> from within an owner thread to release
	 * ownership of the Semaphore.
	 *
	 * @exception	java.lang.IllegalArgumentException
	 *					An IllegalArgumentException will be thrown
	 *					if this method is called within the context
	 *					of a non-owner thread.
	 */
	public void release() {
		synchronized(this) {
			// Get current Thread. If ! an owner, throw exception.
			Thread t = Thread.currentThread();
			if(!_hashOwnership.containsKey(t))
				throw(new IllegalArgumentException(
						"Only an owning Thread may release a Semaphore."));
					
			// Remove owner from list.
			_hashOwnership.remove(t);
		}
		
		// If count < max, make signaled.
		if(_hashOwnership.size() < _maxOwners)
			setState(signaled);
			
		return;
	}
	
	/* *********
	 * void unreserve(Thread t)
	 *
	 * After parent class unreserve(), if count < max, set
	 * state to signaled.
	 *
	 ********* */
	
	void unreserve(Thread t) {
		super.unreserve(t);
		if(_hashOwnership.size() < _maxOwners)
			setState(signaled);
		return;
	}
}

