/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  


/*
 *
 * Copyright (c) 1996 Brian Maso. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and
 * without fee is hereby granted.
 *
 * The Synch object is the base class for all synchronization objects
 * that work with the maso.synch.Scheduler.
 *
 * A working Synch implements four methods:
 * + boolean reserve(): a request to reserve the synchronization object
 *       on behalf of a Thread. Returns true if, at the time of the
 *       call, the synchronization object is signaled and has reserved
 *       internal resources for the Thread. Eg, a Semaphore would
 *       reserve a slot for the Thread. The synchronization object
 *       should change its state to Synch.reserved if claiming the
 *       reservation would change the Synch's state to Synch.unsignaled. A
 *       successful reservation guarantees that gain() will work.
 * + void gain(): A Thread which has previously reserved Synch resources
 *       is now claiming those resources. The Synch's state may change
 *       after gain() is called. For example, a Semaphore may change
 *       state to unsignaled after the N-th Thread has gained ownership.
 *       Eg2, a Mutex will change to unsignaled state after the first
 *       Thread has gained ownership.
 * + void unreserve(): A Thread which previously reserved resources
 *       is giving them up without gaining ownership of the Synch.
 *
 * In addition, each different type of Synch must implement a method
 * to release ownership of the Synch by a Thread. For example,
 * the maso.synch.Mutex class implements a release() method. When the
 * owning Thread calls this method, the Mutex automatically switches
 * to a signaled state and notifies the Scheduler using
 * Scheduler.notifysignaled(). It is imperitive that
 * Scheduler.notifysignaled is called from the owning Thread within
 * a Synch object method when the Synch changes state to signaled.
 *
 * Synch.setState() makes this call to the Scheduler automatically,
 * so use Synch.setState() to set the Synch's internal state, and
 * Synch.getState() to see what the Synch's state is.
 *
 * 3/7/96: Object synchronization problem exists, here's an
 * explanation and the fix. Problem is that the Scheduler's running
 * Thread and an external Thread can easily block each other. What
 * happens is the Scheduler's Thread, which owns the Scheduler's
 * monitor, tries to also get a Synch's monitor. The external
 * Thread, OTOH, owns the Synch's monitor, either through
 * Synch.setState() or Synch.destroy(), and is trying to gain the
 * Scheduler's monitor. Both Threads block each other indefinitely.
 *
 * HOW TO FIX: Follow this basic rule: An external Thread may not
 * attempt to gain the Scheduler's monitor while it already has the
 * Synch's monitor. If this rule is followed, the above mutual
 * blocking condition can not happen. To implement this, just make
 * sure that the external Thread does not own the Synch's monitor
 * when it calls either Scheduler.notifysignaled() or
 * Scheduler.unregister().
 *
 */

package com.radview.util.synch;

import java.util.Hashtable;

/**
 * The Synch class is the base class for all Synch's in the
 * maso.synch package. The Synch class implements the minimum
 * functionality common to all synchronization objects in
 * this package.
 *
 * <P>Threads synchronize themselves on the state of one or more
 * Synch objects in the Scheduler. When the Synch's state is
 * "signaled", synchronized threads are released. Synchronized
 * threads are stalled by the Scheduler while the Synch's state
 * is "unsignaled".
 *
 * <P>Synch objects you will want to use include
 * Signal, AutoSignal, Clock, Timer, Mutex, Semaphore, etc.
 *
 * @see Scheduler
 * @see Signal
 * @see AutoSignal
 * @see Clock
 * @see Timer
 * @see TickingClock
 * @see Mutex
 * @see Semaphore
 * @see CouterSynch
 * @see ThreadSynch
 *
 * @version 1.0
 * @author Brian Maso
 */

public class Synch extends Object {
	/**
	 * "signaled" is one of the possible states of a
	 * Synch object. Threads which are synchronized on
	 * the Synch in the Scheduler are only released when
	 * the Synch's state is "signaled".
	 */
	public final static int signaled = 0;

	/**
	 * "reserved" is one of the possible states of a
	 * Synch object. The reserved state represents an intermediate
	 * state used only by the Scheduler.
	 */
	public final static int reserved = 1;

	/**
	 * "unsignaled" is one of the possible states of a
	 * Synch object. Threads which are synchronized on
	 * the Synch in the Scheduler remain suspended while
	 * the Synch's state is "unsignaled".
	 */
	public final static int unsignaled = 2;

	private int _state;

	protected Hashtable _hashOwnership = new Hashtable();

	protected final static int RESERVATION = 0;
	protected final static int OWNERSHIP   = 1;

	Synch() {
		this(unsignaled);
		return;
	}

	Synch(int state) {
		setState(state);
		Scheduler.register(this);
		return;
	}

	/* *********
	 * public boolean reserve(Thread t)
	 *
	 * Called only by the Scheduler's run() method by the
	 * Scheduler's internal thread, a call to this
	 * method is a request for the Synch to guarantee the Thread
	 * may gain() it at some time in the future. If the Synch
	 * must reserve resources for the Thread, then it should.
	 * Thread state should become reserved if any future reservations
	 * should fail. For an unsignaled Synch, this method should
	 * always fail.
	 *
	 ********* */

	synchronized boolean reserve(Thread t) {
		// If Thread already has a reservation or ownership,
		// then ignore this request, returning true.
		if(null != _hashOwnership.get(t))
			return true;

		// If state is signaled, a simple Synch will return
		// true.
		if(signaled == getState()) {
			_hashOwnership.put(t, new Integer(RESERVATION));
			return true;
		}

		return false;
	}

	/* *********
	 * public void gain(Thread t)
	 *
	 * Request from the Scheduler for ownership of this Synch on
	 * behalf of the Thread. Note that 1) only Threads which have
	 * successfully reserved this Synch will be passed to this
	 * method; and 2) any Thread which has successfully reserved
	 * must be allowed Synch resources.
	 *
	 * This method must be called by the Scheduler's internal
	 * Thread, else there is a serious problem.
	 *
	 ********* */

	synchronized void gain(Thread t) {
		// Make sure the Thread has a reservation.
		Integer r = (Integer)_hashOwnership.get(t);
		if(null == r)
			throw(new IllegalArgumentException(
					"Must have a reservation to gain() ownership!"));
            // If thread already has ownership, return benignly.
		if(OWNERSHIP == r.intValue())
			return;

		// Change Thread's status to owner, then return.
		_hashOwnership.put(t, new Integer(OWNERSHIP));
		return;
	}

	/* *********
	 * void unreserve(Thread t)
	 *
	 * Request made by Scheduler to release any resources associated
	 * with the Thread, which has reserved a spot. Note that
	 * if the Thread is an owner, this method should ignore the
	 * request. If is not a reserver, then an IllegalArgumentException
	 * should be thrown, cause that should never happen. If is
	 * a reserver, then reservation should be removed.
	 *
	 ********* */

	synchronized void unreserve(Thread t) {
		// If is not an owner nor reserver, then an IllegalArgumentException
		// must be thrown.
		Integer r = (Integer)_hashOwnership.get(t);
		if(null == r)
			throw(new IllegalArgumentException(
					"Cannot unreserve unless a reserver or owner!"));
		if(OWNERSHIP == r.intValue())
			return;

		_hashOwnership.remove(t);
		return;
	}

	/* *********
	 * set and get state methods. Simple.
	 ********* */

	/**
	 * Gets the current state of the Synch. The return
	 * value will either be <I>signaled</I> or
	 * <I>unsignaled</I>.
	 */
	public synchronized int getState() {
		return _state;
	}

	protected void setState(int state) {
		int nPrevState;

		// Sychronize on this object's monitor, but make
		// sure the monitor is not owned when
		// Scheduler.notifysignaled() is called to avoid a
		// mutual blocking condition with the Scheduler's
		// internal Thread.
		boolean fBecamesignaled;
		boolean fBecameUnsignaled;
		synchronized(this) {
			if((signaled != state) &&
					(reserved != state) &&
					(unsignaled != state))
				throw(new IllegalArgumentException(
						"Initial state must be signaled, reserved or unsignaled!"));

			// If state changing to signaled, make sure to call
			// Scheduler.notifysignaled().
			nPrevState = _state;
			_state = state;

			fBecamesignaled =
					((signaled == state) && (unsignaled == nPrevState));
			fBecameUnsignaled =
					((unsignaled == state) && (signaled == nPrevState));
		}

		if(fBecamesignaled)
			Scheduler.notifysignaled(this);
		else if(fBecameUnsignaled)
			Scheduler.notifyUnsignaled(this);

		return;
	}

	/**********
	 * void destroy()
	 *
	 * When the Synch is destroyed, it must unregister from the
	 * Scheduler to remove all references to it. This causes
	 * all Threads waiting for the Synch to be restarted with
	 * an ABANDONED return value.
	 *
	 * Note that if you do not destroy a Synch, any Threads waiting
	 * for the Synch will indefinitely wait, and the Synch will
	 * never stop having a reference in the Scheduler, so its
	 * memory will never be released.
	 *
	 **********/

	/**
	 * A Synch must be destroyed so that it can deallocate any
	 * resources it has allocated. It's a good idea to destroy
	 * any Synchs as soon as possible in your applications, as
	 * leftover Synchs can degrade the performance of the Scheduler.
	 * Threads waiting for a Synch which is destroyed are automatically
	 * released from <I>Scheduler.waitFor()</I> with a return value of
	 * <I>Scheduler.ABANDONED</I>.
	 *
	 * @see Scheduler#waitFor
	 */

	public void destroy() {
		Scheduler.unregister(this);
		return;
	}

	public String paramString() {
		return (getState()==unsignaled ? "unsignaled" :
				(getState()==signaled ? "signaled" : "reserved"));
	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(getClass().getName() + "[" +
				paramString());
		buf.append("]");
		return buf.toString();
	}
}
