/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
 
/*
 *
 * Copyright (c) 1996 Brian Maso. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and
 * without fee is hereby granted. 
 * 
 * A ThreadSynch is a synchronization object associated with
 * a particular Thread. When the associated Thread terminates,
 * The ThreadSynch becomes signaled.
 *
 */

package com.radview.util.synch;

/**
 * A ThreadSynch is a Synch which only becomes signaled after
 * a particular thread is no longer active. When used in combination
 * with the Scheduler's OR mode, use of multiple ThreadSynchs allows
 * you to wait for <I>any one of</I> several threads to terminate.
 * Here's an example:
 *
 * <PRE>
 *
 *	public class Example implements Runnable {
 *		Thread[] _athreads = new Thread[100];
 *
 *		public static void main(String[] astrArgs) {
 *			new Example();
 *			return;
 *		}
 *
 *		public Example() {
 *			Vector vectThreadSynchs = new Vector();
 *
 *			for(int ii=0 ; ii<_athreads.length ; ii++) {
 *				_athreads[ii] = new Thread(this);
 *				vectThreadSynchs.addElement(
 *						new ThreadSynch(_athreads[ii]));
 *				_athreads[ii].start();
 *			}
 *
 *			while(0 != vectThreadSynchs.size()) {
 *				int index = Scheduler.waitFor(vectThreadSynchs,
 *						Scheduler.OR);
 *				System.out.println("Thread at index " +
 *						index + " died.");
 *				ThreadSynch ts = vectThreadSynchs.elementAt(index);
 *				ts.destroy();
 *				vectThreadSynchs.removeElementAt(index);
 *			}
 *
 *			System.out.println("All done!");	
 *
 *			return;
 *		}
 *
 *		public void run() {
 *			// Wait a random amount of time before
 *			// dying.
 *			Thread.currentThread().sleep((new Random()).nextLong());
 *			return;
 *		}
 *	}
 *
 * </PRE>
 *
 * Note that this will cause your program to <I>never</I> terminate:
 *
 * <PRE>
 *
 *		Scheduler.waitFor(new ThreadSynch(Thread.currentThread()));
 *
 * @version 1.0
 * @author Brian Maso
 */


public class ThreadSynch extends Synch implements Runnable {
	Thread _threadWaiting = new Thread(this,
			"ThreadSynch-" + __nNextThreadID++);
	Thread _threadWaitFor;
	static int __nNextThreadID = 0;

	/**
	 * Pass the constructor a reference to the thread the ThreadSynch
	 * is to watch. When the thread terminates or is no longer active,
	 * the ThreadSynch will become signaled. The ThreadSynch is
	 * unsignaled until that time.
	 *
	 * @param	threadWaitFor		A reference to the Thread
	 *						the ThreadSynch is supposed to watch.
	 */
	public ThreadSynch(Thread threadWaitFor) {
		super(unsignaled);
		
		_threadWaitFor = threadWaitFor;
		if(null == _threadWaitFor)
			throw(new IllegalArgumentException(
					"Cannot wait for Thread null!"));

		_threadWaiting.setDaemon(true);
		_threadWaiting.start();
		return;
	}
	
	/* *********
	 * void gain(Thread t)
	 *
	 * Since there are no owners of a ThreadSynch, the gaining
	 * Thread must be removed from the ownership list.
	 *
	 ********* */
	
	void gain(Thread t) {
		super.gain(t);
		_hashOwnership.remove(t);
	}
	
	/* *********
	 * void destroy()
	 *
	 * The waiting thread must be interrupted and set to null
	 * so that when it wakes up it knows this object has
	 * been destroyed.
	 *
	 ********* */
	
	/**
	 * As with all Synchs, destroy a ThreadSynch when it is no longer needed.
	 * Note that a ThreadSynch uses a background thread which may not stop
	 * running until some time in the future. This thread will not affect your
	 * program's running or orderly shutdown.
	 */
	public void destroy() {
		synchronized(this) {
			Thread threadTempRef = _threadWaiting;
			_threadWaiting = null;
			threadTempRef.stop(); // stop is deprecated
		}
		super.destroy();
		return;
	}
	
	/* **********
	 * void run()
	 *
	 * _threadWaiting runs through this code. Basically,
	 * _threadWaiting joins _threadWaitFor. If the waiting
	 * thread is interrupted, then we quit. Else, this
	 * ThreadSynch is set to the signaled state.
	 *
	 ********* */
	
	public void run() {
		try {
			_threadWaitFor.join();
		} catch (Exception e) {}
		
		if(null == _threadWaiting)
			return;
			
		setState(signaled);
		return;
	}
}
