/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
#include "wlIdentLocation.h"

#define SEPARATOR '.'
#define NUMBER_SEPARATOR '#'
#define OBJECT_INDEX_SEPARATOR ':'
#define RET_OK 1
#define RET_ERROR -1

//! Ident location element class definition
class CWLIdentLocationElement : public RVCollectable 
{
private:
  int       m_Number;
  RVCString m_Name;
public:
  CWLIdentLocationElement (const int Number, const RVCString& Name);
  virtual ~CWLIdentLocationElement() {}
  int Get (int& Number, RVCString& Name);
};

//! Constructor for holding the number and the name of an identification element
CWLIdentLocationElement::CWLIdentLocationElement (const int Number, const RVCString& Name) 
{
  m_Number = Number;
  m_Name   = Name;
}
//! Method return the number and the name of an identification element
int CWLIdentLocationElement::Get (int& Number, RVCString& Name) 
{
  Number = m_Number;
  Name   = m_Name;
  return (RET_OK);
}

//! Default Ctor - iident location is created with no values in it
//! values can be added later
CWLIdentLocation::CWLIdentLocation () 
{
}
//! Ctor which creates the ident location with a default ident location
//! value - more values can be added later
CWLIdentLocation::CWLIdentLocation (const RVCString& strLocation) 
{
  constructFromStr (strLocation, true);
  m_fullIdentLocation = strLocation;
}
//! CCtor which creates the ident location as copied from ident location
//! value - more values can be added later
CWLIdentLocation::CWLIdentLocation (const CWLIdentLocation& IdentLocation) 
{
  constructFromStr (IdentLocation.m_fullIdentLocation, true);
  m_fullIdentLocation = IdentLocation.m_fullIdentLocation;
}
//! DTor - nothing in paticular to do here...
CWLIdentLocation::~CWLIdentLocation() 
{
  m_Elements.clearAndDestroy();
}
//! Append new level into the Ident location. This can be done or by giving a 
//! new level number and name or by giving a string containing those values
int CWLIdentLocation::appendLevel (const int& Number, const RVCString& Name) 
{
  CWLIdentLocationElement* IdentLocationElement = new CWLIdentLocationElement (Number, Name);
  char      TempStr[20];

  if (!IdentLocationElement)
    return (RET_ERROR);

  if (m_Elements.append (IdentLocationElement)) {
    if (m_fullIdentLocation.length() > 0) {
      m_fullIdentLocation.append(SEPARATOR);
    }
    m_fullIdentLocation.append (Name);
    if (Number != NULL_IDENT_LOCATION_NUMBER) {
      m_fullIdentLocation.append (NUMBER_SEPARATOR);
      itoa( Number, TempStr, 10 );
      m_fullIdentLocation.append (TempStr);
    }
    return (RET_OK);
  }
  else {
    delete IdentLocationElement;
    return (RET_ERROR);
  }
}
int CWLIdentLocation::appendLevel (const RVCString& Value) 
{
  if (constructFromStr (Value, true) == RET_OK) {
    return (RET_OK);
  }
  else {
    return (RET_ERROR);
  }
}
//! Prepend new level into the Ident location. This can be done or by giving a 
//! new level number and name or by giving a string containing those values
int CWLIdentLocation::prependLevel (const int& Number, const RVCString& Name)
{
  CWLIdentLocationElement* IdentLocationElement = new CWLIdentLocationElement (Number, Name);
  char      TempStr[20];

  if (!IdentLocationElement)
    return (RET_ERROR);

  if (m_Elements.prepend (IdentLocationElement)) {
    if (m_fullIdentLocation.length() > 0) {
      m_fullIdentLocation.prepend(SEPARATOR);
    }
    if (Number != NULL_IDENT_LOCATION_NUMBER) {
      itoa( Number, TempStr, 10 );
      m_fullIdentLocation.prepend (TempStr);
      m_fullIdentLocation.prepend (NUMBER_SEPARATOR);
    }
    m_fullIdentLocation.prepend (Name);
    return (RET_OK);
  }
  else {
    delete IdentLocationElement;
    return (RET_ERROR);
  }
}
int CWLIdentLocation::prependLevel (const RVCString& Value) 
{
  if (constructFromStr (Value, false) == RET_OK) {
    return (RET_OK);
  }
  else {
    return (RET_ERROR);
  }
}
//! Private method which insert a new level with values.
//! the method will check that the level which we try to nsert is in the range
//! of the elements in the object
int CWLIdentLocation::insertAt (const int& Level, int& Number, RVCString& Name)
{
  if (Level > Depth () || (Level < 0)) {
    return (RET_ERROR);
  }
  else {
    CWLIdentLocationElement* IdentLocationElement = new CWLIdentLocationElement (Number, Name);
    if (m_Elements.insertAt (Level,IdentLocationElement)) {
      return (RET_OK);
    }
    else {
      delete IdentLocationElement;
      return (RET_ERROR);
    }
  }
}

//! Return the number of levels contined in this ident location
int CWLIdentLocation::Depth () 
{
  return m_Elements.entries();
}
//! Get for a specified level the number(must exists) and the name (if exists)
int CWLIdentLocation::getLevel (const int& Level, int& Number, RVCString& Name) 
{
  try
  {
    if (m_Elements.entries() > 0) {
      CWLIdentLocationElement* IdentLocationElement = (CWLIdentLocationElement*)m_Elements[Level];
      return (IdentLocationElement->Get (Number, Name));
    }
    else {
      Number = NULL_IDENT_LOCATION_NUMBER;
      Name.remove (0);
      return RET_OK;
    }
  }
  catch (...)
  {
    return (RET_ERROR);
  }
}
RVCString CWLIdentLocation::getStr ()
{
  return (m_fullIdentLocation);
}
//! Remove level and then reconstruc the string from the elements
int CWLIdentLocation::removeLevel (const int& Level) 
{
  RVCollectable *rem;
  if ( (rem = m_Elements.removeAt(Level)) != NULL ) {
	delete rem;
    constructFromElements ();
    return (RET_OK);
  }
  else {
    return (RET_ERROR);
  }
}
int CWLIdentLocation::getObjectIndex ()
{
  return (m_ObjectIndex);
}
int CWLIdentLocation::setObjectIndex (const int& objectIndex)
{
  char      TempStr[20];

  m_ObjectIndex = objectIndex;
  if (m_ObjectIndex != NULL_IDENT_LOCATION_NUMBER) {
     m_fullIdentLocation.append (OBJECT_INDEX_SEPARATOR);
     itoa( m_ObjectIndex, TempStr, 10 );
     m_fullIdentLocation.append (TempStr);
  }
  return (RET_OK);
}
int CWLIdentLocation::setObjectIndex (const RVCString& objectIndex)
{
  m_ObjectIndex = atoi ((const char *)(objectIndex));
  m_fullIdentLocation.append (OBJECT_INDEX_SEPARATOR);
  m_fullIdentLocation.append (objectIndex);
  return (RET_OK);
}

bool CWLIdentLocation::isEmpty ()
{
  if (m_Elements.entries () == 0)
    return (true);
  else
    return (false);
}

//! Private method which go over the string and sperate it into parts
//! which will be used to construct the levels
int CWLIdentLocation::constructFromStr (const RVCString& FromStr, bool Append) 
{
  size_t    StartPos = 0;
  size_t    EndPos;
  int       Number;
  RVCString Name;
  int       startLevel = 0;
  RVCString intFromStr (FromStr);

  if (Append) {
    size_t    objectIndex = intFromStr.last (OBJECT_INDEX_SEPARATOR) ;
    if (objectIndex != RV_NPOS) {
      m_ObjectIndex = atoi ((const char *)(RVCString)(intFromStr (objectIndex+1,FromStr.length()-objectIndex-1)));
      intFromStr = intFromStr.remove(objectIndex);
    }
    else {
        m_ObjectIndex = NULL_IDENT_LOCATION_NUMBER;
    }
  }

  EndPos = intFromStr.index (SEPARATOR);
  while (EndPos != RV_NPOS) {
    if (findValues (intFromStr (StartPos,EndPos-StartPos),Number,Name) == RET_OK) {    
      if (Append) {
        appendLevel (Number,Name);
      }
      else {
        insertAt (startLevel,Number,Name);
        ++startLevel;
      }
    }
    StartPos = EndPos+1;
    EndPos   = intFromStr.index (SEPARATOR,StartPos);
  }
  if (StartPos != intFromStr.length()) {
    if (findValues (intFromStr (StartPos,intFromStr.length()-StartPos),Number,Name) == RET_OK) {    
      if (Append)
        appendLevel (Number,Name);
      else
        insertAt (startLevel,Number,Name);
    }
  }
  return (RET_OK);
}
//! Private method which go over the string and sperate it into parts
//! which will be used to construct the levels
int CWLIdentLocation::constructFromElements () 
{
  int       Number;
  RVCString Name;
  char      TempStr[20];

  m_fullIdentLocation.remove(0);
  for (size_t i=0;i<m_Elements.entries();++i) {
    if (i>0) {
      m_fullIdentLocation.append (SEPARATOR);
    }
    if (getLevel (i,Number,Name) == RET_OK) {
      m_fullIdentLocation.append (Name);
      if (Number != NULL_IDENT_LOCATION_NUMBER) {
        m_fullIdentLocation.append (NUMBER_SEPARATOR);
        itoa( Number, TempStr, 10 );
        m_fullIdentLocation.append (TempStr);
      }
    }
  }
  if (m_ObjectIndex != NULL_IDENT_LOCATION_NUMBER) {
    m_fullIdentLocation.append (OBJECT_INDEX_SEPARATOR);
    itoa( m_ObjectIndex, TempStr, 10 );
    m_fullIdentLocation.append (TempStr);
  }
  return (RET_OK);
}

//! Private method which finds the pair of number and name values from a specfied string
int CWLIdentLocation::findValues (const RVCString& FromStr, int& Number, RVCString& Name) 
{
  int NumberPos = FromStr.index (NUMBER_SEPARATOR);
  if (NumberPos == RV_NPOS) {
    Name = FromStr;
    Number = NULL_IDENT_LOCATION_NUMBER;
    return (RET_OK);
  }
  else {
    if (NumberPos != 0) {
      Name = FromStr (0,NumberPos);
    }
    else {
      Name.remove(0);
    }
    // @@Migration - for some reason, with new RW, I had to change the formula here
//    Number = atoi ((const char *)((RVCString)(FromStr(NumberPos+1,FromStr.length()-NumberPos))));     
    Number = atoi ((const char *)((RVCString)(FromStr(NumberPos+1,FromStr.length()-NumberPos-1))));     
    return (RET_OK);
  }
}

//! Operators definitions - TBC (To be completed ...)
CWLIdentLocation& CWLIdentLocation::operator+= (const CWLIdentLocation& IdentLocation)
{
  constructFromStr (IdentLocation.m_fullIdentLocation, true);
  m_fullIdentLocation += IdentLocation.m_fullIdentLocation;
  return (*this);
}
CWLIdentLocation& CWLIdentLocation::operator+= (const RVCString& strLocation)
{
  constructFromStr (strLocation, true);
  m_fullIdentLocation += strLocation;
  return (*this);
}
CWLIdentLocation& CWLIdentLocation::operator+  (const CWLIdentLocation& IdentLocation)
{
  CWLIdentLocation* retIdentLocation = new CWLIdentLocation (*this);
  retIdentLocation ->constructFromStr (IdentLocation.m_fullIdentLocation,true);
  retIdentLocation ->m_fullIdentLocation += IdentLocation.m_fullIdentLocation;
  return (*retIdentLocation);
}
CWLIdentLocation& CWLIdentLocation::operator+  (const RVCString& strLocation)
{
  CWLIdentLocation* retIdentLocation = new CWLIdentLocation (*this);
  retIdentLocation ->constructFromStr (strLocation,true);
  retIdentLocation ->m_fullIdentLocation += strLocation;
  return (*retIdentLocation);
}
CWLIdentLocation& CWLIdentLocation::operator=  (const CWLIdentLocation& IdentLocation)
{
  m_Elements.clearAndDestroy();
  constructFromStr (IdentLocation.m_fullIdentLocation, true);
  m_fullIdentLocation = IdentLocation.m_fullIdentLocation;
  return (*this);
}
CWLIdentLocation& CWLIdentLocation::operator=  (const RVCString& strLocation)
{
  m_Elements.clearAndDestroy();
  constructFromStr (strLocation, true);
  m_fullIdentLocation = strLocation;
  return (*this);
}
bool CWLIdentLocation::operator== (const CWLIdentLocation& IdentLocation)
{
  if (m_fullIdentLocation == IdentLocation.m_fullIdentLocation) {
    return (true);
  }
  else {
    return (false);
  }
}
bool CWLIdentLocation::operator== (const RVCString& strLocation)
{
  if (m_fullIdentLocation == strLocation) {
    return (true);
  }
  else {
    return (false);
  }
}
