/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
// WLSslCiphersCollection.cpp: implementation of the CWLSslCiphersCollection class.
//
//////////////////////////////////////////////////////////////////////
#include <openssl\ssl.h>

#include "connpool.h"
#include "commandmanager.h"
#include "workspace.h"
#include "httpdefines.h"
#include "WLSslCiphersCollection.h"
#include "WLSecureSslConnection.h"
#include "WLGenericConnection.h"

#define reportError m_Cp->ReportError

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWLSslCiphersCollection::CWLSslCiphersCollection( CWLConnectionsPool* cp )
: m_Method( NULL )
, m_Context( NULL )
, m_Ssl( NULL )
, m_Init( false )
, m_Count( 0 )
, m_Strength( 0 )
, m_SslVersion( CWLSecureSslConnection::SSL_Version_Undetermined )
, m_IDs( NULL )
, m_numOfCiphersInList(0)
{
	m_Cp = cp;

	m_SCCLock = new RVCriticalSection();
}


CWLSslCiphersCollection::~CWLSslCiphersCollection()
{
	if( m_Ssl ){
		SSL_free( m_Ssl );
		m_Ssl = NULL;
	}
	if( m_Context ){
		SSL_CTX_free( m_Context );
		m_Context = NULL;
	}

	m_Cp = NULL;

	m_Init = false;
	m_Count = 0;
	m_Strength = 0;
	
	if (m_SCCLock)
		delete m_SCCLock;
}


//////////////////////////////////////////////////////////////////////
// Public methods
//////////////////////////////////////////////////////////////////////

int CWLSslCiphersCollection::Init( int strength, bool verifyStrength )
{
	int err = 1;

	if( InitTemplate() )
		return 0;

	if (m_Init) {
		if (strength != DEFAULT_ENCRYPTION_STRENGTH &&
				strength != m_RequestedStrength)
			reportError(CIPHER_ERROR_FILE, 50, m_Strength);
		return err;
	}

	m_RequestedStrength = strength;

	m_Init = true;

	if (strength == DEFAULT_ENCRYPTION_STRENGTH)	{ // this is the default, so do nothing (2^32 - 1)
		err = 1;
		strength = m_Strength;	// here m_Strength is the max
	}

	else if (strength == 0) {
		//resetAll();
		err = 1;
	}

	else if (strength > 0) {
		//resetAll();
		err = SetStrength(strength);
	}

	return err;
}


int CWLSslCiphersCollection::DoCommand( int command, const char* cipherName )
{
	int retStatus = 1;
	bool wasChanged = false;
	/*
	possible commands:
	all:                 enables all ciphers
	none:                disables all ciphers
	enable <cipher>:     enables the specified cipher
	disable <cipher>:    disables the specified cipher
	showall:             displays all known ciphers
	showallenabled:      displays all enabled ciphers
	*/

	if ( command == CIPHER_ENABLE_ALL ) {
		retStatus = EnableAll(wasChanged);
	}
	else if ( command == CIPHER_DISABLE_ALL ) {
		retStatus = DisableAll(wasChanged);
	}
	else if ( command == CIPHER_ENABLE ) {
		EnableNamed(cipherName, wasChanged);
	}
	else if ( command == CIPHER_DISABLE ) {
		DisableNamed(cipherName, wasChanged);
	}
	else if ( command == CIPHER_SHOW_ALL ) {
		ShowAll();
	}
	else if ( command == CIPHER_SHOW_ENABLED ) {
		ShowEnabled();
	}
  return (retStatus);
}


int CWLSslCiphersCollection::SetStrength( int s )
{
		return 0;
}


void CWLSslCiphersCollection::EnableId( int index, bool& wasChanged )
{
	wasChanged = false;
}


void CWLSslCiphersCollection::DisableId( int index, bool& wasChanged )
{
	wasChanged = false;
}


void CWLSslCiphersCollection::EnableNamed( const char* cipherName, bool& wasChanged )
{
	wasChanged = false;
}


void CWLSslCiphersCollection::DisableNamed( const char* cipherName, bool& wasChanged )
{
	wasChanged = false;
}


int CWLSslCiphersCollection::NameToId( const char * name )
{
	if( NULL == name || NULL == m_Method )
		return -1;

	STACK_OF(SSL_CIPHER) * ciphers_stack = SSL_get_ciphers( m_Ssl );
	if( NULL == ciphers_stack ){
		return -1;
	}

	int i = 0;
	SSL_CIPHER* cipher = NULL;

	do{
		cipher = sk_SSL_CIPHER_value( ciphers_stack, i++ );
		if( NULL == cipher )
			break;

		if( 0 == strcmp( cipher->name, name ) )
			return cipher->id;

	}while( true );
	
	return -1;
}


const char * CWLSslCiphersCollection::IdToName( int index )
{
	if( NULL == m_Method )
		return NULL;
	
	STACK_OF(SSL_CIPHER) * ciphers_stack = SSL_get_ciphers( m_Ssl );
	if( NULL == ciphers_stack ){
		return NULL;
	}
	
	int i = 0;
	SSL_CIPHER* cipher = NULL;
	
	do{
		cipher = sk_SSL_CIPHER_value( ciphers_stack, i++ );
		if( NULL == cipher )
			break;
		
		if( cipher->id == index )
			return cipher->name;
		
	}while( true );

	return NULL;
}


int CWLSslCiphersCollection::GetCipherCount()
{
	if( NULL == m_Method )
		return 0;

	int count =0;
	STACK_OF(SSL_CIPHER) * ciphers_stack = SSL_get_ciphers( m_Ssl );
	if( NULL == ciphers_stack ){
		return 0;
	}

	int i = 0;
	do{
		SSL_CIPHER* cipher = sk_SSL_CIPHER_value( ciphers_stack, i++ );
		if( cipher ){
			count++;
		}
		else break;	
	}while( true );

	return count;
}


void CWLSslCiphersCollection::GetStrengthList(char** list)
{
	*list = NULL;

	if( NULL == m_Ssl )
		return;

	STACK_OF( SSL_CIPHER ) * stack = SSL_get_ciphers( m_Ssl );
	if( NULL == stack )
		return;

	int numOfCiphers = getEnabledCiphersCount( stack );;
	int* intList = (int*)calloc(numOfCiphers, sizeof(int));

	if (NULL == intList)
		return;

	int emptySlot = 0;

	for( int i = 0; i < numOfCiphers; i++ ){
		SSL_CIPHER* cr = sk_SSL_CIPHER_value( stack, i );
		if( cr ){
			if ( cr->strength_bits == 0 || ( i > 0 && isInList( cr->strength_bits, intList, emptySlot ) ) )
				continue;
			else
				intList[emptySlot++] = cr->strength_bits;
		}
	}

	// allocate 8 bytes for each strength found
	*list = (char*)calloc(emptySlot * 8, sizeof(char));
	if (!list) {
		free(intList);
		return;
	}

	bool first = true;
	char num[16];
	strcat(*list, "0,");
	for ( i = 0; i < emptySlot; i++ ){
		if (!first)
			strcat(*list, ",");
		else
			first = false;
		sprintf(num, "%d", intList[i]);
		strcat(*list, num);
	}

	free(intList);
}


bool CWLSslCiphersCollection::Update( CWLGenericConnection* pConnection )
{
	RVCriticalSection::LockGuard Lock(*m_SCCLock);
	
	bool retCode = false;

	if( NULL == pConnection )
			return false;

	retCode = pConnection->SetCiphers( m_Strength );

	return retCode;
}


//////////////////////////////////////////////////////////////////////
// Private methods
//////////////////////////////////////////////////////////////////////
int CWLSslCiphersCollection::InitTemplate()
{
	if( NULL != m_Method )
		if(  m_Cp->GetCm()->GetWorkspace()->m_SslVersion == m_SslVersion )
			return 1;
		if( m_Ssl ){
			SSL_free( m_Ssl );
			m_Ssl = NULL;
		}
		if( m_Context ){
			SSL_CTX_free( m_Context );
			m_Context = NULL;
		}

	m_SslVersion = m_Cp->GetCm()->GetWorkspace()->m_SslVersion;

	// set the client method according to SslVersion
	if( NULL == m_Method )
		switch( m_SslVersion ){
		case SSL_Version_2_0:
			m_Method = SSLv2_client_method();
			break;

		case SSL_Version_3_0_Only:
		case SSL_Version_3_0:
			m_Method = SSLv3_client_method();
			break;

		case TLS_Version_1_0_With_2_0_Hello:
		case TLS_Version_1_0_Only:
		case TLS_Version_1_0:
			m_Method = TLSv1_client_method();
			break;

		default:
			m_Method = SSLv23_client_method();
		}
		
	if( NULL == m_Method )
		return 0;


	// create the context
	if( NULL == m_Context )
		m_Context = SSL_CTX_new( m_Method );
	if( NULL == m_Context ){
		m_Method = NULL;
		return 0;
	}

	// tune the context
	switch( m_SslVersion ){
	case SSL_Version_3_0_Only:
		SSL_CTX_set_options( m_Context, SSL_OP_NO_SSLv2 | SSL_OP_NO_TLSv1  );
		break;
	case TLS_Version_1_0_Only:
		SSL_CTX_set_options( m_Context, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3 );
		break;
	}

	if( NULL == m_Ssl )
		m_Ssl = SSL_new( m_Context );
	if( NULL == m_Ssl ){
		m_Method = NULL;
		SSL_CTX_free( m_Context );
		m_Context = NULL;
		return 0;
	}

	int strength = m_Cp->GetCm()->GetWorkspace()->m_SslCryptoStrength;
	RVCString strStrength;

	switch( strength ){
	case SSL_AllCrypto:
		strStrength = "ALL:eNULL";
		break;
	case SSL_ExportCryptoOnly:
		strStrength = "EXPORT";
		break;
	case SSL_StrongCryptoOnly:
		strStrength = "HIGH";
		break;
	default:
		strStrength = "DEFAULT";
// what about SSL_ServerGatedCrypto ????
	}

	if( ! SSL_set_cipher_list( m_Ssl, strStrength.data() ) )
		return 0;

	return 1;
}


int CWLSslCiphersCollection::EnableAll( bool& wasChanged )
{
	int retStatus = 0;
	int i = 0;
	wasChanged = false;
	
	if( NULL == m_Ssl )
		return retStatus;
	
	if( !SSL_set_cipher_list( m_Ssl, "ALL:eNULL" ) )
		return 0;

	wasChanged = true;

	STACK_OF(SSL_CIPHER)* stack = SSL_get_ciphers( m_Ssl );
	if( NULL == stack )
		return 0;

	while( true ){
		SSL_CIPHER* cr = sk_SSL_CIPHER_value( stack, i++ );
		if( NULL != cr ){
			WlTraceMsg("* Enabled cipher %s [KeySize:%4d] (SSL Id:0x%0.4hX)", cr->name, cr->strength_bits, cr->id);
		}
		else
			return 1;
	}

	return 1;
}


int CWLSslCiphersCollection::DisableAll( bool& wasChanged )
{
	int i = 0;

	if( NULL == m_Ssl)
		return 0;

	STACK_OF(SSL_CIPHER)* stack = SSL_get_ciphers( m_Ssl );
	if( NULL == stack )
		return 0;

	do{
		SSL_CIPHER* cr = sk_SSL_CIPHER_value( stack, i++ );
		if( NULL != cr ){
			WlTraceMsg("* Removed cipher %s [KeySize:%4d] (SSL Id:0x%0.4hX)", cr->name, cr->strength_bits, cr->id);
		}
		else{
			break;
		}
	}while( true );

	SSL_set_cipher_list( m_Ssl, "" );

	wasChanged = true;

	return 1;
}


void CWLSslCiphersCollection::ShowAll()
{
	if( NULL == m_Method )
		return;

	int numOfCiphers = m_Method->num_ciphers();

	if( 0 == numOfCiphers ){
		WlTraceMsg("* There are not any known ciphers (%d):", numOfCiphers);
		return;
	}

	WlTraceMsg("* Show all known ciphers (%d):", numOfCiphers);

	for( int i = 0; i < numOfCiphers; i++ ){
		SSL_CIPHER* cr =  m_Method->get_cipher( i );
		if( cr ){
			WlTraceMsg("Name: %s, SSL Id:0x%0.4hX : ", cr->name, cr->id);}
	}
}


void CWLSslCiphersCollection::ShowEnabled()
{
	int numOfCiphers = 0;
	
	if( NULL == m_Ssl )
		return;

	// get the cipher stack
	STACK_OF(SSL_CIPHER)* stack = SSL_get_ciphers( m_Ssl );

	// count enabled ciphers
	if( NULL == stack ){
		return;
	}

	numOfCiphers = getEnabledCiphersCount( stack );

	if( 0 == numOfCiphers ){
		WlTraceMsg("* There are not any enabled ciphers (%d):", numOfCiphers);
		return;
	}

	// show enabled ciphers
	WlTraceMsg("* Show all enabled ciphers (%d):", numOfCiphers);
	numOfCiphers = 0;

	do{
		SSL_CIPHER* cr = sk_SSL_CIPHER_value( stack, numOfCiphers++ );
		if( cr ){
			WlTraceMsg("Name: %s, SSL Id:0x%0.4hX : ", cr->name, cr->id);
			reportError( CIPHER_ERROR_FILE, 80, cr->name, cr->id );
		}
		else break;
	}while( true );
}


bool CWLSslCiphersCollection::isInList( int bits, int* list, int listLength )
{
	for (int i = 0; i < listLength; i ++) {
		if (list[i] == bits)
			return true;
	}

	return false;
}


int CWLSslCiphersCollection::getEnabledCiphersCount( void* ciphersStack )
{
	int numOfCiphers = 0;

	if( NULL == m_Ssl)
		return 0;
	
	STACK_OF(SSL_CIPHER)* stack = (STACK_OF(SSL_CIPHER)*) ciphersStack;

	do{
		SSL_CIPHER* cr = sk_SSL_CIPHER_value( stack, numOfCiphers );
		if( NULL == cr )
			break;

		numOfCiphers++;
	}while( true );

	return numOfCiphers;
}
