/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
#include "isession.h"
#include "associativearray.h"
#include "wlleprotocol.h"
#include "wldlls.h"
#include "loadengineinifileproperties.h"
#include "loadengineselistener.h"


iSession::iSession(
	CGeneralProcess& rProcess )
:
	mProcess(rProcess),
	m_WarningCounter(0),
	m_ErrorCounter(0),
	m_CompletedRounds(0)
{
}

iSession::~iSession()
{
}


CGeneralProcess& iSession::fGetProcess()
{
	return mProcess;
}

double iSession::fGetTimeFromBegin() {
	double rtn(0);
	double aBeginTime( fGetBeginTime() );
  if( 0 != aBeginTime ) {
    rtn = mProcess.fGetCurrentTime() - aBeginTime;
	}
	return rtn;
}

iLEProtocolContainer& iSession::GetLEProtocolProcessContainer() {
	return fGetProcess().GetLEProtocolProcessContainer();
}

iLEProtocolContainer& iSession::GetLEProtocolSessionContainer() {
	return m_ProtocolSessionContainer;
}

bool iSession::InitProtocols(const char* pDisabledProtocols[]) {

	bool ret(true);
  WLProtocolIndex indx;

  iLEProtocolContainer& a_ProtocolProcessContainer = fGetProcess().GetLEProtocolProcessContainer();

  WLProtocolIndex a_NumProtocols = a_ProtocolProcessContainer.NumProtocols();

  m_ProtocolSessionContainer.InitProtocolContainer(a_NumProtocols);

	WlTraceMsg("Calling NewProtocolSession for all protocols");
  for (indx = 0; indx < a_NumProtocols; indx++){
    if (a_ProtocolProcessContainer.ProtocolExists(indx)){
      WlAssert(dynamic_cast<iLEProtocolProcess*>(&(a_ProtocolProcessContainer.GetProtocolAtIndex(indx))));
      iLEProtocolProcess& a_ProtocolProcess = static_cast<iLEProtocolProcess&>(a_ProtocolProcessContainer.GetProtocolAtIndex(indx));
      bool Disable = false;
      if (pDisabledProtocols && *pDisabledProtocols){
        RVCString currentDLLName = static_cast<cDll*>(fGetProcess().GetDllProtocols()[indx])->GetNameDll();
        for (int i = 0; pDisabledProtocols[i]; i++){
          if (currentDLLName.compareTo(CLEIniFileProperties::GetAppendProtocolRealName(pDisabledProtocols[i]), RVCString::ignoreCase) == 0){
            Disable = true;
            break;
          }
        }
      }
      if (!Disable){
        iLEProtocolSession& a_ProtocolSession= a_ProtocolProcess.NewProtocolSession();
        a_ProtocolSession.Init(this);
        m_ProtocolSessionContainer.InsertProtocolAtIndex(indx, a_ProtocolSession);
      }
    }
  }

	WlTraceMsg("Calling SessionInitialize for all protocols");
  for (indx = 0; indx < a_NumProtocols; indx++){
    if (m_ProtocolSessionContainer.ProtocolExists(indx)){
      WlAssert(dynamic_cast<iLEProtocolSession*>(&(m_ProtocolSessionContainer.GetProtocolAtIndex(indx))));
      if (!static_cast<iLEProtocolSession&>(m_ProtocolSessionContainer.GetProtocolAtIndex(indx)).SessionInitialize()){
				WlTraceMsg("Error 5 in function iSession::InitProtocols");
				ret = false;
      }
    }
  }

	return ret;
}

void iSession::ProtocolSessionTermination(){
	WlTraceMsg("Calling SessionTermination for all protocols");
	for (WLProtocolIndex indx = m_ProtocolSessionContainer.NumProtocols() - 1; indx >= 0 ; indx--) {
    if (m_ProtocolSessionContainer.ProtocolExists(indx)){
      WlAssert(dynamic_cast<iLEProtocolSession*>(&(m_ProtocolSessionContainer.GetProtocolAtIndex(indx))));
      if (!static_cast<iLEProtocolSession&>(m_ProtocolSessionContainer.GetProtocolAtIndex(indx)).SessionTermination()){
    		WlTraceMsg("Error 6 in function iSession::ProtocolSessionTermination");
      }
    }
	}
}

bool iSession::fInit(const char* pDisabledProtocols[]){

	try{

		LE_SET_SE_TRANSLATOR();

		return InitProtocols(pDisabledProtocols);

	}

  catch (std::exception* e) {
		ExceptionHandler(e);
		delete e; // Only for catch of pointer 
		return false;
  }
  catch (std::exception& e) {
		ExceptionHandler(&e);
		return false;
  }
  catch (...) {
    WlErrorMsg("Internal Error : Unknown occurred in iSession::fInit");
		return false;
	}	

}

void iSession::fFinal(){
  ProtocolSessionTermination();
}

void iSession::IncrementErrorCounter() {
	if (CLEIniFileProperties::ErrorCountersOn){
		RVMutexLock::LockGuard lock(m_ErrorsLock);
		if (m_ErrorCounter == INT_MAX)
			m_ErrorCounter = 0;
		else
			++m_ErrorCounter;
	}
}

void iSession::IncrementWarningCounter() {
	if (CLEIniFileProperties::ErrorCountersOn){
		RVMutexLock::LockGuard lock(m_ErrorsLock);
		if (m_ErrorCounter == INT_MAX)
			m_WarningCounter = 0;
		else
			++m_WarningCounter;
	}
}

unsigned long iSession::GetCompletedRounds(){
	RVMutexLock::LockGuard lock(m_CompletedRoundsLock);
	return m_CompletedRounds;
}

void iSession::IncrementCompletedRounds(){
	RVMutexLock::LockGuard lock(m_CompletedRoundsLock);
	++m_CompletedRounds;
}


void iSession::ExceptionHandler(std::exception* e){
	RVCString FullExcName ("");
	WLERRGetErrorMsg(e, FullExcName);
	RVCString StackTrace("");
	if (WLERRGetStackTrace(e, StackTrace) && StackTrace.length() > 0){
		WlErrorMsg("iSession::fInit - Exception %s occured\nTrace:\n%s", FullExcName, StackTrace);
		printf("iSession::fInit - Exception %s occured\nTrace:\n%s", FullExcName, StackTrace);
	}
	else {
		WlErrorMsg("iSession::fInit - Exception %s occured", FullExcName);
		printf("iSession::fInit - Exception %s occured", FullExcName);
	}
}
