/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
// WLStatGenMeasurementDictionary.cpp

#include <math.h>
#include <thr.h>

#include "tracedll.h"
#include "wllightstring.h"
#include "tvpnumconversions.h"
#include "wlstatisticsgenerator.h"
#include "wlstatmanagerglobals.h"
 

/////////////////////////////////////////////////////////////////
///                        CWLStatGenMeasurementDictionary
/////////////////////////////////////////////////////////////////


CWLStatGenMeasurementDictionary::CWLStatGenMeasurementDictionary(CWLStatGenSlice *pCS) : 
  RVHashDictionary(33), 
  pSlice(pCS),
  pMyClient(NULL)
{
  // Chaim Apr. 16 2001 - Initing stats cache variables
  m_stat_counter = 0;
  mutex = new RVTRecursiveLock<RVMutexLock>;
}

CWLStatGenMeasurementDictionary::~CWLStatGenMeasurementDictionary() 
{ 
  if (pMyClient) {
    pMyClient->ReleaseDictionary();
    pMyClient = NULL;
  }
  pSlice = NULL;
  clearAndDestroy(); 
  delete mutex;
}

void CWLStatGenMeasurementDictionary::Print()
{
  RVTRecursiveLock<RVMutexLock>::LockGuard lock(*mutex);
  RVHashDictionaryIterator hdi((RVHashDictionary &) *this);
  while (hdi()){
    cout << ((CWLCollectableString *) hdi.key())->data() << ":";
    ((CWLStatGenMeasurement *) hdi.value())->Print();
  }
}

void CWLStatGenMeasurementDictionary::ResetLasts()
// Only Used to ResetLasts at the history Slice dictionaries.
// This is the reason why we don't need to Lock it.
{
  RVHashDictionaryIterator hdi((RVHashDictionary &) *this);
  while (hdi()){
    ((CWLStatGenMeasurement *) hdi.value())->ResetLasts();
  }
}


CWLStatGenMeasurement  *CWLStatGenMeasurementDictionary::GetMeasurement(const char *name)
{
  CWLCollectableString k(name, false);
  CWLStatGenMeasurement *sb;
  sb= (CWLStatGenMeasurement *) findValue((RVCollectable *) &k); 
  return sb;
}


void CWLStatGenMeasurementDictionary::SetValues(
               const char *name, 
               double min, double max, int cnt, double s,double sd,
               double l_min, double l_max, int l_cnt, double l_sum, 
               CWLStatPointTypeEnum::PointType type)
{
  CWLCollectableString k(name, false);
  CWLStatGenMeasurement *sb;

  RVTRecursiveLock<RVMutexLock>::LockGuard lock(*mutex);
	if (GetFlag()==false) 
    return; // The dictionary cannot be use, since it was packed
  
	if ((sb = (CWLStatGenMeasurement *) findValue((RVCollectable *) &k)) == NULL)
	{
    sb = new CWLStatGenMeasurement(type);
    insertKeyAndValue((RVCollectable *) new CWLCollectableString(name), 
			                (RVCollectable *) sb);
  }
  WlTraceMsg(" ----------------- ");
  WlTraceMsg(" CWLStatGenMeasurementDictionary::SetValues ");
  sb->SetValues(min, max, cnt, s, (s*s), l_min, l_max, l_cnt, l_sum, (l_sum*l_sum), type);
  WlTraceMsg("Sum =  %d     SumSqr = %d ", (int)s, (int)(s*s));
  WlTraceMsg(" ----------------- ");
	
	return;
}

// Chaim Apr. 16 2001 - Flush stats cache.
bool CWLStatGenMeasurementDictionary::FlushStatsCache ()
{

  WlTraceMsg("In FlushStatsCache");
  if (m_stat_counter == 0){
    WlTraceMsg("Nothing to flush");
    return true;
  }

  RVTRecursiveLock<RVMutexLock>::LockGuard lock(*mutex);
  int i;
  for (i = 0; i < m_stat_counter; i++){
    CWLCollectableString k(m_stat[i].name, false);
    CWLStatGenMeasurement *sb;
    WlTraceMsg("Adding stat: %s=%f", m_stat[i].name, m_stat[i].l);

    if ((sb = (CWLStatGenMeasurement *) findValue((RVCollectable *) &k)) == NULL)
  	{
      sb = new CWLStatGenMeasurement(m_stat[i].type);
      insertKeyAndValue((RVCollectable *) new CWLCollectableString(m_stat[i].name), 
		  	                (RVCollectable *) sb);
    }
    sb->AddLast(m_stat[i].l);
  }
  m_stat_counter = 0;

  return true;
}

// Chaim Apr. 16 2001 - Added cache.
// The cache is flushed either when cache is full, every 5 seconds and
// it's also called by the load engine before entering sleep.

bool CWLStatGenMeasurementDictionary::AddLast(const char *name, double l, CWLStatPointTypeEnum::PointType type)
{
  // Chaim - Moved back from FlushStatsCache()
  if (GetFlag()==false) { 
    return false; // The dictionary cannot be use, since it was packed
  }

//  if (m_stat_counter == 0) m_last_update_time = TVPSysTime();

  WlTraceMsg("Caching stat: %s=%f", name, l);
  
  // Truncate string
  if (strlen(name) <= MAX_STAT_NAME_LEN){
    strcpy (m_stat[m_stat_counter].name, name);
  }
  else {
    strncpy (m_stat[m_stat_counter].name, name, MAX_STAT_NAME_LEN);
    m_stat[m_stat_counter].name[MAX_STAT_NAME_LEN] = '\0';
  }

  m_stat[m_stat_counter].l = l;
  m_stat[m_stat_counter].type = type;
  m_stat_counter++;

  // Chaim - In probing client, this will always be null and we should just flush
  if (pSlice){
    if (m_stat_counter < pSlice->GetGenerator()->m_stat_cache_size)
      return true;

    if (m_stat_counter >= pSlice->GetGenerator()->m_stat_cache_size) {
      WlTraceMsg("Flushing stats due to cache full");
    }
    else {
      WlTraceMsg("Flushing stats due to timeout");
    }
  }

  return FlushStatsCache ();

}


void CWLStatGenMeasurementDictionary::Reset(const char *name)
{
  CWLCollectableString k(name, false);
  CWLStatGenMeasurement *sb;

  RVTRecursiveLock<RVMutexLock>::LockGuard lock(*mutex);
  sb = (CWLStatGenMeasurement *) findValue( (RVCollectable *) &k);
	if ( sb!=NULL)  sb->Reset();
}


void CWLStatGenMeasurementDictionary::AddValuesOfDict(CWLStatGenMeasurementDictionary *from,AddType type)
{
  CWLCollectableString *k;
  CWLStatGenMeasurement *sbFrom, *sbTo;

  RVTRecursiveLock<RVMutexLock>::LockGuard lock(*(from->mutex));
  RVHashDictionaryIterator hdi((RVHashDictionary &) *from);
  while (hdi()){
    k = (CWLCollectableString *) hdi.key();
    sbFrom = (CWLStatGenMeasurement *) hdi.value();
		if (sbFrom==NULL) return;
		sbTo = (CWLStatGenMeasurement *) findValue((RVCollectable *) k);
		if (sbTo == NULL)
		{
      sbTo = new CWLStatGenMeasurement(sbFrom->GetType());
      insertKeyAndValue((RVCollectable *) new CWLCollectableString(k->data()), 
											  (RVCollectable *) sbTo);
		}	
		sbTo->AddValues(sbFrom->GetMinimum(), sbFrom->GetMaximum(),
		            sbFrom->GetCount(), sbFrom->GetSum(), sbFrom->GetSumSqr(),
								sbFrom->GetLastsMinimum(),sbFrom->GetLastsMaximum(), 
								sbFrom->GetLastsCount(),sbFrom->GetLastsSum(), sbFrom->GetLastsSumSqr(), type);			
		
	}
}


void CWLStatGenMeasurementDictionary::DumpDict(RVCString &str)
{
  RVTRecursiveLock<RVMutexLock>::LockGuard lock(*mutex);
  str = "";
  RVHashDictionaryIterator hdi((RVHashDictionary &) *this);
  while (hdi()){
		 CWLCollectableString *k = (CWLCollectableString *) hdi.key();
     CWLStatGenMeasurement *sb = (CWLStatGenMeasurement *) hdi.value();
		 sb->AppendDump(str, k->data());
    }
}

void CWLStatGenMeasurementDictionary::DumpLasts(RVCString &str)
{
  RVTRecursiveLock<RVMutexLock>::LockGuard lock(*mutex);
  str = "";
  RVHashDictionaryIterator hdi((RVHashDictionary &) *this);
  while (hdi()){
		 CWLCollectableString *k = (CWLCollectableString *) hdi.key();
     CWLStatGenMeasurement *sb = (CWLStatGenMeasurement *) hdi.value();
		 sb->DumpLasts(str, k->data());
    }
}


double CWLStatGenMeasurementDictionary::GetMinimum(const char *name)
{
	return GetMeasurement(name)->GetMinimum();
}

double CWLStatGenMeasurementDictionary::GetMaximum(const char *name)
{
	// -1 flag Remark the measurement is not exist
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetMaximum();
	else
		return -1;

}

int    CWLStatGenMeasurementDictionary::GetCount(const char *name)
{
	// -1 flag Remark the measurement is not exist
	if (GetMeasurement(name)!=NULL)
		return (int) GetMeasurement(name)->GetCount();
	else
		return -1;
}

double  CWLStatGenMeasurementDictionary::GetSum(const char *name)
{
		// -1 flag Remark the measurement is not exist
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetSum();
	else
		return -1;

}

double  CWLStatGenMeasurementDictionary::GetSumSqr(const char *name)
{
		// -1 flag Remark the measurement is not exist
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetSumSqr();
	else
		return -1;
}

double  CWLStatGenMeasurementDictionary::GetStddev(const char *name)
{
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetStddev();
	else
		return -1;
}

double  CWLStatGenMeasurementDictionary::GetAverage(const char *name)
{	
	if (GetMeasurement(name)!=NULL)
			return GetMeasurement(name)->GetAverage();
	else
		return -1;
}


double  CWLStatGenMeasurementDictionary::GetLastsMinimum(const char *name)
{
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetLastsMinimum();
	else
		return -1;
}

double  CWLStatGenMeasurementDictionary::GetLastsMaximum(const char *name)
{
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetLastsMaximum();
	else
		return -1;
}

int     CWLStatGenMeasurementDictionary::GetLastsCount(const char *name) 
{
	if (GetMeasurement(name)!=NULL)
		return (int) GetMeasurement(name)->GetLastsCount();
	else
		return -1;
}

double  CWLStatGenMeasurementDictionary::GetLastsSum(const char *name)	 
{
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetLastsSum();
	else
		return -1;
}  

double  CWLStatGenMeasurementDictionary::GetLastsSumSqr(const char *name)	 
{
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetLastsSumSqr();
	else
		return -1;
}  

double  CWLStatGenMeasurementDictionary::GetLastsAverage(const char *name)
{
	if (GetMeasurement(name)!=NULL)
		return GetMeasurement(name)->GetLastsAverage();
	else
		return -1;
}

