/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
// UDPSender.cpp: implementation of the UDPSender class.
//
//////////////////////////////////////////////////////////////////////

#include "udpsender.h"
#include "tracedll.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

UDPSender::UDPSender()
{
	m_Running = m_Initialized = false;
	m_SenderThreadEvent.Reset();
// BEGIN: create a functor
#if !defined(RV_THR_CANT_EXTRACT_TEMPLATE_ARGS_FROM_SIGNATURE)
	m_SenderThread = rvMakeThreadFunction(
														 *this,
														 &UDPSender::ThreadLoop);
#else
	m_SenderThread = rvMakeThreadFunctionM(
														 UDPSender,
														 *this, void,
											    	 &UDPSender::ThreadLoop);
#endif
// END: create a functor
}

UDPSender::~UDPSender()
{
	m_Running = m_Initialized = false;
}

int UDPSender::Init(u_short port, const char* hostName)
{
	int status;
	int enable = 1;
#ifdef __TV_WINDOWS__
  WSADATA Data;
#endif

	if(m_Initialized)
		return true;

  /// initialize the Windows Socket DLL
#ifdef __TV_WINDOWS__
  status = WSAStartup(MAKEWORD(1, 1), &Data);
  if(status != 0)
    WlTraceMsg("ERROR: UDPSender::Init - WSAStartup unsuccessful");
#endif

	// specify the IP address
	if(!UpdateToHost(hostName))
		return false;
  // specify the port portion of the address
  m_DestSockAddr.sin_port = htons(port);
  // specify the address family as Internet
  m_DestSockAddr.sin_family = AF_INET;

  // create a socket
  m_DestSocket = socket(AF_INET, SOCK_DGRAM, 0);
  if(m_DestSocket == INVALID_SOCKET) {
    WlTraceMsg("ERROR: UDPSender::Init - socket unsuccessful");
#ifdef __TV_WINDOWS__
    status = WSACleanup();
    if(status == SOCKET_ERROR)
      WlTraceMsg("ERROR: UDPSender::Init - WSACleanup unsuccessful");
#endif
    return false;
  }

  // permit broadcasting on the socket
  status = setsockopt(m_DestSocket,
											SOL_SOCKET,
											SO_BROADCAST,
											(char *) &enable,
											sizeof(enable));
  if(status != 0)
    WlTraceMsg("ERROR: UDPSender::Init - setsockopt unsuccessful");

	m_Initialized = true;

	return true;
}

int UDPSender::Start(const char* message, long duration)
{
	int startStatus;
	m_Running = true;
	m_MessageToSend = message;
	m_timeVal = duration;

	// start the thread and wait for it's execution:
	startStatus = m_SenderThread.start();
	m_SenderThread.wait(RV_THR_RUNNING);

	return startStatus;
}

void UDPSender::ThreadLoop(void)
{
	// BEGIN: Set priority
	{
		RVThreadSelf self = ::rvThread();
		self.setPriority(self.getMaxPriority());
	}
	// END: Set priority

	while(m_Running) {
		Run(); // Run returns false incase of an error
		m_SenderThreadEvent.Wait(m_timeVal);
	}
}

// Default Run method
int UDPSender::Run(void)
{
	if( !SendToHost(m_MessageToSend.data()) ) {
		WlTraceMsg("ERROR: EASTestTalkClient::Run - failed sending message(%s)", m_MessageToSend.data());
		return false;
	}
	return true;
}

void UDPSender::Stop(void)
{
	if(!m_Running) {
		if(m_Initialized)
			Terminate();
		return;
	}

	m_Running = false;
	// terminate net conection
	if(m_Initialized)
		Terminate();

	m_SenderThreadEvent.Set();

	// wait for thread to exit for 5 sec' and terminate
	// it if it didn't succeed to exit by it self.
	if(m_SenderThread.join(5000) == RV_THR_TIMEOUT) {
		m_SenderThread.terminate();
	}
	//Terminate();
}

int UDPSender::SendToHost(const char* buffer)
{
  int numSent;

  numSent = sendto(m_DestSocket, buffer,
								   strlen(buffer) + 1, NO_FLAGS_SET,
									 (struct sockaddr *)&m_DestSockAddr,
									 sizeof(m_DestSockAddr));

  if(numSent != (int)strlen(buffer) + 1) {
    return false;
  }
	return true;
}

int UDPSender::SendTo(const char* hostName, const char* buffer)
{
  int status = false;

	if(CSocketLib::GetHostByName(hostName, m_DestSockAddr.sin_addr)) {
		status = SendToHost(buffer);
		CSocketLib::GetHostByName(m_HostName.data(), m_DestSockAddr.sin_addr);
	}
	return status;
}

int UDPSender::UpdateToHost(const char* hostName)
{
	int status = false;

	if(hostName == NULL) {
		char localHostName[MAXBUFLEN];
		if(CSocketLib::GetLocalHostName(localHostName, MAXBUFLEN)) {
			if(CSocketLib::GetHostByName(localHostName, m_DestSockAddr.sin_addr)) {
				status = true;
				m_HostName = localHostName;
			}
		}
	}
	else {
		m_HostName = hostName;
		status = CSocketLib::GetHostByName(m_HostName.data(), m_DestSockAddr.sin_addr);
	}
	return status;
}

bool UDPSender::SenderStatusRunning(void)
{
	return m_Running;
}

void UDPSender::Terminate(void)
{
	int status;

	m_Initialized = false;

#ifdef __TV_WINDOWS__
	status = closesocket(m_DestSocket);
#else
	status = close(m_DestSocket);
#endif
  if(status == SOCKET_ERROR)
		WlTraceMsg("ERROR: UDPSender::Terminate - closesocket unsuccessful");

#ifdef __TV_WINDOWS__
  status = WSACleanup();
  if(status == SOCKET_ERROR)
    WlTraceMsg("ERROR: UDPSender::Terminate - WSACleanup unsuccessful");
#endif
}


