/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
// errdef.h -- error mechanism definitions
//
//   who          when          what
//  Iris        28-11-93      Created
//
//
#ifndef __WLERRORLIB_H
#define __WLERRORLIB_H

#include <RVCString.h>
#include <exception>
#include "platform.h"

//class DllExport RVCString;

class DllExport WLLogMessage {
public:
  typedef enum {
	  Success,
	  Note,
	  Warning,
	  Error,
    Fatal	
  } Severity;

  static const char* Name(Severity severity);
  static const char* LongName(Severity severity);
};

class DllExport WLErrorsDictionary {
public:
  void * dict;

public:
  WLErrorsDictionary();
  virtual ~WLErrorsDictionary();

  int LoadErrorString(const char* errFileName, const char* data);

  const char* GetErrorFormat(const char* errFileName, int fmtId);
  WLLogMessage::Severity GetErrorSeverity(const char* errFileName, int fmtId);
  const char* GetErrorString(const char* errFileName, int fmtId, char* errBuf, unsigned int errBufLen, ...);
private:
  int LoadErrorFile(const char* errFileName);
};

class DllExport iWLException : public std::exception {
public:
    enum {
        DEFAULT_STACKTRACE_DEPTH = 100
    };

	iWLException();
	virtual ~iWLException() RV_EXCEPTION_THROW;
	
	virtual const char* what() const RV_EXCEPTION_THROW; // Override standard default behavior
	virtual const char* GetErrorMsg() const = 0;

    // Description:
    //   obtain the stack trace from the given exception
    // 
    // Parameters
    //   skip - the number of stack frames skipped at the beginning
    //   fmt - stack trace format, similar to printf format spec where the accepted codes are:
    //     %m : module name
    //     %f : filename
    //     %l : line number
    //     %s : symbol (function name)
    //   maxStackDepth - the maximum number of stackframes included in the stacktrace (0 means unlimited)
    //
    // Returns:
    //   True if successful

	virtual bool StackTrace(RVCString& outBuf, unsigned skip = 0, const char* fmt = 0, unsigned int maxStackDepth = DEFAULT_STACKTRACE_DEPTH) = 0;
};

class DllExport WLErrorException : public iWLException {
private:
  RVCString errorMsg;
  WLLogMessage::Severity severity;
  bool alreadyReported;

public:
  WLErrorException(const char* msg, WLLogMessage::Severity sev, ...);
  WLErrorException(bool reported, const char* msg, WLLogMessage::Severity sev, ...);
  WLErrorException(const char* errFileName, int fmtId, ...);
  WLErrorException(bool reported, const char* errFileName, int fmtId, ...);
  // @@ofer: the format of copy ctor should be X(const X&) but since WebRM 4.6 uses the same dlls we cannot modify the interface
  // in 6.0 we should use X(const X&) for ALL platforms
#ifdef __TV_UNIX__
  WLErrorException(const WLErrorException& errorException);
#else
  WLErrorException(WLErrorException& errorException);
#endif // __TV_UNIX__
  virtual ~WLErrorException() RV_EXCEPTION_THROW;

  virtual const char* GetErrorMsg() const;
  WLLogMessage::Severity GetErrorSeverity() const;
  bool GetReported() { return alreadyReported;}
  virtual bool StackTrace(RVCString& outBuf, unsigned skip = 0, const char* fmt = 0, unsigned int maxStackDepth = iWLException::DEFAULT_STACKTRACE_DEPTH);
};

class DllExport WLErrorExceptionWithStackTrace : public WLErrorException {
private:
  RVCString m_stackTrace;

public:

  WLErrorExceptionWithStackTrace(const char* msg, WLLogMessage::Severity sev, unsigned int maxDepth = iWLException::DEFAULT_STACKTRACE_DEPTH);
  // @@ofer: the format of copy ctor should be X(const X&) but since WebRM 4.6 uses the same dlls we cannot modify the interface
  // in 6.0 we should use X(const X&) for ALL platforms
#ifdef __TV_UNIX__
  WLErrorExceptionWithStackTrace(const WLErrorExceptionWithStackTrace& errorException);
#else
  WLErrorExceptionWithStackTrace(WLErrorExceptionWithStackTrace& errorException);
#endif // __TV_UNIX__
  virtual ~WLErrorExceptionWithStackTrace() RV_EXCEPTION_THROW;

  virtual bool StackTrace(RVCString& outBuf, unsigned skip = 0, const char* fmt = 0, unsigned int maxStackDepth = iWLException::DEFAULT_STACKTRACE_DEPTH);

private:
	void SetStackTrace(unsigned int maxStackDepth = iWLException::DEFAULT_STACKTRACE_DEPTH);
};

class DllExport CWLErrorInfo {
public:
	void Clear();

  CWLErrorInfo(const char* p_cErrFileName="")
    : m_nErrorSeverity(WLLogMessage::Success),
		  m_nErrorId(0),
      m_sErrorFile(p_cErrFileName)
  {

  }
  
  virtual ~CWLErrorInfo(){;}


  void SetError(int p_nErrorId, ...);

  const char* GetErrorMsg();
  WLLogMessage::Severity GetErrorSeverity();
  int GetErrorId() {return m_nErrorId;}

  void SetErrorMsg(const char* p_sErrorMsg){m_nErrorId = 0; m_sErrorMsg=p_sErrorMsg;}
  void SetErrorSeverity(WLLogMessage::Severity p_nErrorSeverity){ m_nErrorSeverity=p_nErrorSeverity;}
  void SetErrorId(int p_nErrorId){ m_sErrorMsg=""; m_nErrorId = p_nErrorId;}

private:
  RVCString               m_sErrorFile;
  RVCString               m_sErrorMsg;
  WLLogMessage::Severity  m_nErrorSeverity;
  int                     m_nErrorId;

};


DllExport int WLERRLoadErrorString(const char* errFileName, const char* data);
DllExport const char* WLERRGetErrorFormat(const char* errFileName, int fmtId);
DllExport const char* WLERRGetifErrorFormat(const char* errFileName, int fmtId);
DllExport WLLogMessage::Severity WLERRGetErrorSeverity(const char* errFileName, int fmtId);
DllExport const char* WLERRGetErrorString(const char* errFileName, int fmtId, char* errBuf, unsigned int errBufLen, ...);
DllExport void WLERRGetErrorMsg(std::exception*, RVCString&);
DllExport bool WLERRGetStackTrace(std::exception*, RVCString&);
DllExport void WLERRTerminate();

#endif // __WLERRORLIB_H

