/* WebLOAD Open Source is an automated quality assurance testing tool for internet applications.
 * Copyright (C) 1996 Radview Software, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * only under the terms of Version 2 of the GNU General Public License as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; Without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; If not, you can download it from 
 * http://www.gnu.org/licenses/
 *
 */
  
#ifndef __TV_UNIX__
#include <crtdbg.h>
#endif
#include <collint.h>


#include "tvppriorityqueue.h"

class CPriorityQueueItem : public RVCollectableInt {
private:
  RVCollectable   *m_Item;

public:
  CPriorityQueueItem(RVCollectable *item, int priority) : 
    RVCollectableInt(priority), m_Item(item) 
  {
  }

  virtual ~CPriorityQueueItem() 
  { 
    if (m_Item)
      delete m_Item;
  }

  RVCollectable *GetCollectable() 
  {
    RVCollectable *temp = m_Item;
    m_Item = NULL;
    return temp;
  }
};


TVPPriorityQueue::TVPPriorityQueue() 
{
  m_Q = new RVSortedVector;
  m_Touch = new RVMutexLock;
  m_Condition  = new RVCondition(*m_Touch);
  Init();
}

TVPPriorityQueue::~TVPPriorityQueue() 
{
  Terminate();
  delete m_Q;
  delete m_Touch;
  delete m_Condition;
}

void TVPPriorityQueue::Init()
{
  RVMutexLock::LockGuard lock(*m_Touch);
  m_Released = false;
  m_Active = true;
}

void TVPPriorityQueue::Terminate()
{
  RVMutexLock::LockGuard lock(*m_Touch);
  m_Active = false;
  m_Released = true;
  m_Q->clearAndDestroy();
	m_Condition->signalAll();
}

void TVPPriorityQueue::release()
{
  RVMutexLock::LockGuard lock(*m_Touch);
  m_Released = true;
	m_Condition->signalAll();
}

void TVPPriorityQueue::putEvent(RVCollectable* event, int priority)
{
  RVMutexLock::LockGuard lock(*m_Touch);
  if (!m_Active || m_Released)
		return;
  CPriorityQueueItem *newItem = new CPriorityQueueItem(event, priority);
	m_Q->insert((RVCollectable *)newItem);
	m_Condition->signal();
}

RVCollectable* TVPPriorityQueue::waitEvent()
{
  RVMutexLock::LockGuard lock(*m_Touch);

	//
	// wait for the message
  while (m_Active && !m_Released && m_Q->isEmpty()) {
  	m_Condition->wait();
  }

  if (!m_Active || (m_Released && m_Q->isEmpty()))
		return NULL;

  CPriorityQueueItem *item = (CPriorityQueueItem*)m_Q->removeAt(0);
	RVCollectable* e = item->GetCollectable();
  delete item;
	return e;
}
	
TVPPriorityQueue::WaitReturnType TVPPriorityQueue::waitEvent(RVCollectable*& event, long waitTime)
{
  RVMutexLock::LockGuard lock(*m_Touch);
  event = NULL;

	//
	// wait for the message
  while (m_Active && !m_Released && m_Q->isEmpty()) {
  	RVWaitStatus status = m_Condition->wait(waitTime);
    if (status == RV_THR_TIMEOUT)
      return TimedOut;
  }

  if (!m_Active || (m_Released && m_Q->isEmpty())) {
    return Released;
  }

  CPriorityQueueItem *item = (CPriorityQueueItem*)m_Q->removeAt(0);
	event = item->GetCollectable();
  delete item;
	return Success;
}

bool TVPPriorityQueue::isEmpty() const
{ 
  RVMutexLock::LockGuard lock(*m_Touch);
  return m_Q->isEmpty(); 
}

